# -----------------------------------------------------------------------------
# Copyright (c) 2020- Spyder Project Contributors
#
# Released under the terms of the MIT License
# (see LICENSE.txt in the project root directory for details)
# -----------------------------------------------------------------------------

"""
Spyder API toolbar widgets.
"""

from __future__ import annotations

# Standard library imports
import os
import sys
import uuid
from collections import OrderedDict
from typing import Literal, Union, TYPE_CHECKING

if sys.version_info < (3, 10):
    from typing_extensions import TypeAlias
else:
    from typing import TypeAlias  # noqa: ICN003

# Third part imports
from qtpy.QtCore import QEvent, QObject, QSize, Qt, Signal
from qtpy.QtWidgets import (
    QAction,
    QProxyStyle,
    QStyle,
    QToolBar,
    QToolButton,
    QWidget,
)

# Local imports
import spyder.utils.qthelpers  # For fully-qualified SpyderAction in type alias
from spyder.api.exceptions import SpyderAPIError
from spyder.api.translations import _
from spyder.api.widgets.menus import SpyderMenu, SpyderMenuProxyStyle
from spyder.utils.icon_manager import ima
from spyder.utils.qthelpers import SpyderAction
from spyder.utils.stylesheet import (
    APP_TOOLBAR_STYLESHEET,
    PANES_TOOLBAR_STYLESHEET,
)

if TYPE_CHECKING:
    from qtpy.QtWidgets import QMainWindow, QStyleOption


# Generic type aliases
ToolbarItem: TypeAlias = Union[spyder.utils.qthelpers.SpyderAction, QWidget]
"""Type alias for the set of supported objects that can be toolbar items."""

ToolbarItemEntry: TypeAlias = tuple[
    ToolbarItem, Union[str, None], Union[str, None], Union[str, None]
]
"""Type alias for the full tuple entry in the list of toolbar items."""


# ---- Constants
# ----------------------------------------------------------------------------
class ToolbarLocation:
    """Pseudo-enum listing possible locations for a toolbar."""

    Top: Qt.ToolBarArea = Qt.TopToolBarArea
    """Toolbar at the top of the layout."""

    Bottom: Qt.ToolBarArea = Qt.BottomToolBarArea
    """Toolbar at the bottom of the layout."""


# ---- Event filters
# ----------------------------------------------------------------------------
class _ToolTipFilter(QObject):
    """
    Filter tooltip events on toolbar buttons.
    """

    def eventFilter(self, obj: QObject, event: QEvent) -> bool:
        """
        Filter tooltip events on toolbar buttons.

        Parameters
        ----------
        obj : QObject
            The object receiving the event.
        event : QEvent
            The event object.

        Returns
        -------
        bool
            ``True`` the event should be filtered out, ``False`` otherwise.
        """
        event_type = event.type()
        action = obj.defaultAction() if isinstance(obj, QToolButton) else None
        if event_type == QEvent.ToolTip and action is not None:
            if action.tip is None:
                return action.text_beside_icon

        return QObject.eventFilter(self, obj, event)


# ---- Styles
# ----------------------------------------------------------------------------
class ToolbarStyle(QProxyStyle):
    """Proxy style class to control the style of Spyder toolbars.

    .. deprecated:: 6.2

        This class will be renamed to the private :class:`!_ToolbarStyle`
        in Spyder 6.2, while the current public name will become an alias
        raising a :exc:`DeprecationWarning` on use, and removed in 7.0.

        It was never intended to be used directly by plugins, and its
        functionality is automatically inherited by using the appropriate
        :class:`ApplicationToolbar` and :class:`MainWidgetToolbar` classes.
    """

    TYPE: Literal["Application"] | Literal["MainWidget"] | None = None
    """
    The toolbar type; must be either "Application" or "MainWidget".
    """

    def pixelMetric(
        self, pm: QStyle.PixelMetric, option: QStyleOption, widget: QWidget
    ) -> int:
        """
        Adjust size of toolbar extension button (in pixels).

        From `Stack Overflow <https://stackoverflow.com/a/27042352/438386>`__.

        This is a callback intended to be called internally by Qt.

        Parameters
        ----------
        pm : QStyle.PixelMetric
            The pixel metric to calculate.
        option : QStyleOption | None, optional
            The current style options, or ``None`` (default).
        widget : QWidget | None, optional
            The widget the pixel metric will be used for,
            or ``None`` (default).

        Returns
        -------
        int
            The resulting pixel metric value, used internally by Qt.

        Raises
        ------
        SpyderAPIError
            If :attr:`TYPE` is not ``"Application"`` or ``"MainWidget"``,
            as then this style would do nothing.
        """
        # Important: These values need to be updated in case we change the size
        # of our toolbar buttons in utils/stylesheet.py. That's because Qt only
        # allow to set them in pixels here, not em's.
        if pm == QStyle.PM_ToolBarExtensionExtent:
            if self.TYPE == "Application":
                if os.name == "nt":
                    return 40
                elif sys.platform == "darwin":
                    return 54
                else:
                    return 57
            elif self.TYPE == "MainWidget":
                if os.name == "nt":
                    return 36
                elif sys.platform == "darwin":
                    return 42
                else:
                    return 44
            else:
                raise SpyderAPIError(
                    "Toolbar style must be 'Application' or 'MainWidget', not"
                    f" {self.TYPE!r}"
                )
        return super().pixelMetric(pm, option, widget)


# ---- Toolbars
# ----------------------------------------------------------------------------
class SpyderToolbar(QToolBar):
    """
    This class provides toolbars with some predefined functionality.

    .. caution::

        This class isn't intended to be used directly; use its subclasses
        :class:`ApplicationToolbar` and :class:`MainWidgetToolbar` instead.
    """

    sig_is_rendered: Signal = Signal()
    """
    Signal to let other objects know that the toolbar is now rendered.
    """

    def __init__(self, parent: QWidget | None, title: str) -> None:
        """
        Create a new toolbar object.

        Parameters
        ----------
        parent : QWidget | None
            The parent widget of this one, or ``None``.
        title : str
            The localized title of this toolbar, to display in the UI.

        Returns
        -------
        None
        """
        super().__init__(parent=parent)

        # Attributes
        self._title = title
        self._section_items = OrderedDict()
        self._item_map: dict[str, ToolbarItem] = {}
        self._pending_items: dict[str, list[ToolbarItemEntry]] = {}
        self._default_section = "default_section"
        self._filter = None

        self.setWindowTitle(title)

        # Set attributes for extension button.
        # From https://stackoverflow.com/a/55412455/438386
        ext_button = self.findChild(QToolButton, "qt_toolbar_ext_button")
        ext_button.setIcon(ima.icon("toolbar_ext_button"))
        ext_button.setToolTip(_("More"))

        # Set style for extension button menu (not all extension buttons have
        # it).
        if ext_button.menu():
            ext_button.menu().setStyleSheet(
                SpyderMenu._generate_stylesheet().toString()
            )

            ext_button_menu_style = SpyderMenuProxyStyle(None)
            ext_button_menu_style.setParent(self)
            ext_button.menu().setStyle(ext_button_menu_style)

    def add_item(
        self,
        action_or_widget: ToolbarItem,
        section: str | None = None,
        before: str | None = None,
        before_section: str | None = None,
        omit_id: bool = False,
    ) -> None:
        """
        Add action or widget item to the given toolbar ``section``.

        Parameters
        ----------
        action_or_widget: ToolbarItem
            The item to add to the toolbar.
        section: str | None, optional
            The section id in which to insert the ``action_or_widget``,
            or ``None`` (default) for no section.
        before: str | None, optional
            Make the ``action_or_widget`` appear before the action with the
            identifier ``before``. If ``None`` (default), add it to the end.
            If ``before`` is not ``None``, ``before_section`` will be ignored.
        before_section : str | None, optional
            Make the ``section`` appear prior to ``before_section``.
            If ``None`` (the default), add the section to the end.
            If you provide a ``before`` action, the new action will be placed
            before this one, so the section option will be ignored, since the
            action will now be placed in the same section as the ``before``
            action.
        omit_id: bool, optional
            If ``False``, the default, then the toolbar will check if
            ``action.action_id`` exists and is set to a string, and raise
            an :exc:`~spyder.api.exceptions.SpyderAPIError` if either is not
            the case. If ``True``, it will add the ``action_or_widget`` anyway.

        Returns
        -------
        None

        Raises
        ------
        SpyderAPIError
            If ``omit_id`` is ``False`` (the default) and
            ``action_or_widget.action_id`` does not exist or is not set to
            a string.
        """
        item_id = None
        if isinstance(action_or_widget, SpyderAction) or hasattr(
            action_or_widget, "action_id"
        ):
            item_id = action_or_widget.action_id
        elif hasattr(action_or_widget, "ID"):
            item_id = action_or_widget.ID
        if not omit_id and item_id is None and action_or_widget is not None:
            raise SpyderAPIError(
                f"Item {action_or_widget} must declare an ID attribute."
            )

        if before is not None:
            if before not in self._item_map:
                before_pending_items = self._pending_items.get(before, [])
                before_pending_items.append(
                    (action_or_widget, section, before, before_section)
                )
                self._pending_items[before] = before_pending_items
                return
            else:
                before = self._item_map[before]

        if section is None:
            section = self._default_section

        action_or_widget._section = section

        if before is not None:
            if section == self._default_section:
                action_or_widget._section = before._section
                section = before._section

        if section not in self._section_items:
            self._section_items[section] = [action_or_widget]
        else:
            if before is not None:
                new_actions_or_widgets = []
                for act_or_wid in self._section_items[section]:
                    if act_or_wid == before:
                        new_actions_or_widgets.append(action_or_widget)
                    new_actions_or_widgets.append(act_or_wid)

                self._section_items[section] = new_actions_or_widgets
            else:
                self._section_items[section].append(action_or_widget)
        if (
            before_section is not None
            and before_section in self._section_items
        ):
            new_sections_keys = []
            for sec in self._section_items.keys():
                if sec == before_section:
                    new_sections_keys.append(section)
                if sec != section:
                    new_sections_keys.append(sec)
            self._section_items = OrderedDict(
                (section_key, self._section_items[section_key])
                for section_key in new_sections_keys
            )

        if item_id is not None:
            self._item_map[item_id] = action_or_widget
            if item_id in self._pending_items:
                item_pending = self._pending_items.pop(item_id)
                for item, section, before, before_section in item_pending:
                    self.add_item(
                        item,
                        section=section,
                        before=before,
                        before_section=before_section,
                    )

    def remove_item(self, item_id: str) -> None:
        """
        Remove the toolbar item with the given string identifier.

        Parameters
        ----------
        item_id : str
            The string identifier of the toolbar item to remove.

        Returns
        -------
        None
        """
        try:
            item = self._item_map.pop(item_id)
            for section in list(self._section_items.keys()):
                section_items = self._section_items[section]
                if item in section_items:
                    section_items.remove(item)
                if len(section_items) == 0:
                    self._section_items.pop(section)
            self.clear()
            self.render()
        except KeyError:
            pass

    def render(self) -> None:
        """
        Render the toolbar taking into account sections and locations.

        Returns
        -------
        None
        """
        sec_items = []
        for sec, items in self._section_items.items():
            for item in items:
                sec_items.append([sec, item])

            sep = QAction(self)
            sep.setSeparator(True)
            sec_items.append((None, sep))

        if sec_items:
            sec_items.pop()

        for sec, item in sec_items:
            if isinstance(item, QAction):
                add_method = super().addAction
            else:
                add_method = super().addWidget

            add_method(item)

            if isinstance(item, QAction):
                widget = self.widgetForAction(item)

                if self._filter is not None:
                    widget.installEventFilter(self._filter)

                text_beside_icon = getattr(item, "text_beside_icon", False)
                if text_beside_icon:
                    widget.setToolButtonStyle(Qt.ToolButtonTextBesideIcon)

                if item.isCheckable():
                    widget.setCheckable(True)

        self.sig_is_rendered.emit()


class ApplicationToolbar(SpyderToolbar):
    """
    A Spyder main application toolbar.

    These toolbars are placed above all Spyder dockable plugins in the
    interface.
    """

    ID: str | None = None
    """
    Unique string toolbar identifier.

    This is used by Qt to be able to save and restore the state of widgets.
    """

    def __init__(
        self, parent: QMainWindow, toolbar_id: str, title: str
    ) -> None:
        """
        Create a main Spyder application toolbar.

        Parameters
        ----------
        parent : QMainWindow
            The parent main window of this toolbar.
        toolbar_id : str
            The unique string identifier of this toolbar.
        title : str
            The localized name of this toolbar, displayed in the interface.

        Returns
        -------
        None
        """
        super().__init__(parent=parent, title=title)
        self.ID = toolbar_id

        self._style = ToolbarStyle(None)
        self._style.TYPE = "Application"
        self._style.setParent(self)
        self.setStyle(self._style)

        self.setStyleSheet(str(APP_TOOLBAR_STYLESHEET))

    def __str__(self) -> str:
        """
        Output this toolbar's class name and identifier as a string.

        Returns
        -------
        str
            The toolbar's class name and string identifier, in the format
            :file:``ApplicationToolbar({TOOLBAR_ID})``.
        """
        return f"ApplicationToolbar('{self.ID}')"

    def __repr__(self) -> str:
        """
        Output this toolbar's class name and identifier as a string.

        Returns
        -------
        str
            The menu's class name and string identifier, in the format
            :file:``ApplicationToolbar({TOOLBAR_ID})``.
        """
        return f"ApplicationToolbar('{self.ID}')"


class MainWidgetToolbar(SpyderToolbar):
    """
    A Spyder dockable plugin toolbar.

   This is used by dockable plugins to have their own toolbars.
    """

    ID: str | None = None
    """
    Unique string toolbar identifier.
    """

    def __init__(
        self, parent: QWidget | None = None, title: str | None = None
    ) -> None:
        """
        Create a new toolbar.

        Parameters
        ----------
        parent : QWidget | None, optional
            The parent widget of this one, or ``None`` (default).
        title : str | None, optional
            The localized title of this toolbar, or ``None`` (default) for
            no title.

        Returns
        -------
        None
        """
        super().__init__(parent, title=title or "")
        self._icon_size = QSize(16, 16)

        # Setup
        self.setObjectName(
            "main_widget_toolbar_{}".format(str(uuid.uuid4())[:8])
        )
        self.setFloatable(False)
        self.setMovable(False)
        self.setContextMenuPolicy(Qt.PreventContextMenu)
        self.setIconSize(self._icon_size)

        self._style = ToolbarStyle(None)
        self._style.TYPE = "MainWidget"
        self._style.setParent(self)
        self.setStyle(self._style)

        self.setStyleSheet(str(PANES_TOOLBAR_STYLESHEET))

        self._filter = _ToolTipFilter()

    def set_icon_size(self, icon_size: QSize) -> None:
        """
        Set the icon size for this toolbar.

        Parameters
        ----------
        icon_size : QSize
            The icon size to set.

        Returns
        -------
        None
        """
        self._icon_size = icon_size
        self.setIconSize(icon_size)



# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
# (see spyder/__init__.py for details)

"""
Default layout definitions.
"""

# Local imports
from spyder.api.plugins import Plugins
from spyder.api.translations import _
from spyder.plugins.layout.api import BaseGridLayoutType


class DefaultLayouts:
    SpyderLayout = "Spyder Default Layout"
    HorizontalSplitLayout = "Horizontal split"
    VerticalSplitLayout = "Vertical split"
    RLayout = "Rstudio layout"
    MatlabLayout = "Matlab layout"


class SpyderLayout(BaseGridLayoutType):
    ID = DefaultLayouts.SpyderLayout

    def __init__(self, parent_plugin):
        super().__init__(parent_plugin)

        self.add_area(
            [Plugins.Projects, Plugins.OutlineExplorer],
            row=0,
            column=0,
            row_span=2,
            visible=False,
        )
        self.add_area(
            [Plugins.Editor],
            row=0,
            column=1,
            row_span=2,
        )
        self.add_area(
            [
                Plugins.Help,
                Plugins.VariableExplorer,
                Plugins.Debugger,
                Plugins.Profiler,
                Plugins.Plots,
                Plugins.Explorer,
                Plugins.Find,
                Plugins.OnlineHelp,
            ],
            row=0,
            column=2,
            default=True,
            hidden_plugin_ids=[Plugins.OnlineHelp, Plugins.Find]
        )
        self.add_area(
            [Plugins.IPythonConsole, Plugins.History, Plugins.Console],
            row=1,
            column=2,
            hidden_plugin_ids=[Plugins.Console]
        )

        self.set_column_stretch(0, 1)
        self.set_column_stretch(1, 3)
        self.set_column_stretch(2, 3)

    def get_name(self):
        return _("Default layout")


class HorizontalSplitLayout(BaseGridLayoutType):
    ID = DefaultLayouts.HorizontalSplitLayout

    def __init__(self, parent_plugin):
        super().__init__(parent_plugin)

        self.add_area(
            [Plugins.Editor],
            row=0,
            column=0,
        )
        self.add_area(
            [
                Plugins.IPythonConsole,
                Plugins.Explorer,
                Plugins.Help,
                Plugins.VariableExplorer,
                Plugins.Debugger,
                Plugins.Profiler,
                Plugins.Plots,
                Plugins.History,
            ],
            row=0,
            column=1,
            default=True,
        )

        self.set_column_stretch(0, 5)
        self.set_column_stretch(1, 4)

    def get_name(self):
        return _("Horizontal split")


class VerticalSplitLayout(BaseGridLayoutType):
    ID = DefaultLayouts.VerticalSplitLayout

    def __init__(self, parent_plugin):
        super().__init__(parent_plugin)

        self.add_area(
            [Plugins.Editor],
            row=0,
            column=0,
        )
        self.add_area(
            [
                Plugins.IPythonConsole,
                Plugins.Explorer,
                Plugins.Help,
                Plugins.VariableExplorer,
                Plugins.Debugger,
                Plugins.Profiler,
                Plugins.Plots,
                Plugins.History,
            ],
            row=1,
            column=0,
            default=True,
        )

        self.set_row_stretch(0, 6)
        self.set_row_stretch(1, 4)

    def get_name(self):
        return _("Vertical split")


class RLayout(BaseGridLayoutType):
    ID = DefaultLayouts.RLayout

    def __init__(self, parent_plugin):
        super().__init__(parent_plugin)

        self.add_area(
            [Plugins.Editor],
            row=0,
            column=0,
        )
        self.add_area(
            [Plugins.IPythonConsole, Plugins.Console],
            row=1,
            column=0,
            hidden_plugin_ids=[Plugins.Console]
        )
        self.add_area(
            [
                Plugins.VariableExplorer,
                Plugins.Debugger,
                Plugins.Profiler,
                Plugins.Plots,
                Plugins.History,
                Plugins.OutlineExplorer,
                Plugins.Find,
            ],
            row=0,
            column=1,
            default=True,
            hidden_plugin_ids=[Plugins.OutlineExplorer, Plugins.Find]
        )
        self.add_area(
            [Plugins.Explorer, Plugins.Projects, Plugins.Help,
             Plugins.OnlineHelp],
            row=1,
            column=1,
            hidden_plugin_ids=[Plugins.Projects, Plugins.OnlineHelp]
        )

    def get_name(self):
        return _("Rstudio layout")


class MatlabLayout(BaseGridLayoutType):
    ID = DefaultLayouts.MatlabLayout

    def __init__(self, parent_plugin):
        super().__init__(parent_plugin)

        self.add_area(
            [Plugins.Explorer, Plugins.Projects],
            row=0,
            column=0,
            hidden_plugin_ids=[Plugins.Projects]
        )
        self.add_area(
            [Plugins.OutlineExplorer],
            row=1,
            column=0,
        )
        self.add_area(
            [Plugins.Editor],
            row=0,
            column=1,
        )
        self.add_area(
            [Plugins.IPythonConsole, Plugins.Console],
            row=1,
            column=1,
            hidden_plugin_ids=[Plugins.Console]
        )
        self.add_area(
            [
                Plugins.VariableExplorer,
                Plugins.Debugger,
                Plugins.Profiler,
                Plugins.Plots,
                Plugins.Find,
            ],
            row=0,
            column=2,
            default=True,
            hidden_plugin_ids=[Plugins.Find]
        )
        self.add_area(
            [Plugins.History, Plugins.Help, Plugins.OnlineHelp],
            row=1,
            column=2,
            hidden_plugin_ids=[Plugins.OnlineHelp]
        )

        self.set_column_stretch(0, 2)
        self.set_column_stretch(1, 3)
        self.set_column_stretch(2, 2)

        self.set_row_stretch(0, 3)
        self.set_row_stretch(1, 2)

    def get_name(self):
        return _("Matlab layout")


class VerticalSplitLayout2(BaseGridLayoutType):
    ID = "testing layout"

    def __init__(self, parent_plugin):
        super().__init__(parent_plugin)

        self.add_area([Plugins.IPythonConsole], 0, 0, row_span=2)
        self.add_area([Plugins.Editor], 0, 1, col_span=2)
        self.add_area([Plugins.Explorer], 1, 1, default=True)
        self.add_area([Plugins.Help], 1, 2)
        self.add_area([Plugins.Console], 0, 3, row_span=2)
        self.add_area(
            [Plugins.VariableExplorer], 2, 0, col_span=4, visible=False)

        self.set_column_stretch(0, 1)
        self.set_column_stretch(1, 4)
        self.set_column_stretch(2, 4)
        self.set_column_stretch(3, 1)

        self.set_row_stretch(0, 2)
        self.set_row_stretch(1, 2)
        self.set_row_stretch(2, 1)

    def get_name(self):
        return _("testing layout")


if __name__ == "__main__":
    for layout in [
            # SpyderLayout(None),
            # HorizontalSplitLayout(None),
            # VerticalSplitLayout(None),
            # RLayout(None),
            # MatlabLayout(None),
            VerticalSplitLayout2(None),
            ]:
        layout.preview_layout(show_hidden_areas=True)



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright (c) 2009- Spyder Project Contributors
#
# Distributed under the terms of the MIT License
# (see spyder/__init__.py for details)
# -----------------------------------------------------------------------------

"""
Tours Container.
"""

from collections import OrderedDict

# Local imports
from spyder.api.exceptions import SpyderAPIError
from spyder.api.translations import _
from spyder.api.widgets.main_container import PluginMainContainer
from spyder.plugins.tours.tours import TourIdentifiers
from spyder.plugins.tours.widgets import AnimatedTour, OpenTourDialog


# Set the index for the default tour
DEFAULT_TOUR = TourIdentifiers.IntroductionTour


class TourActions:
    """
    Tours actions.
    """
    ShowTour = "show tour"


class ToursContainer(PluginMainContainer):
    """
    Tours container.
    """

    def __init__(self, name, plugin, parent=None):
        super().__init__(name, plugin, parent=parent)

        self._main = plugin.main
        self._tours = OrderedDict()
        self._tour_titles = OrderedDict()
        self._tour_widget = AnimatedTour(self._main)

        self._tour_dialog = OpenTourDialog(
            self, lambda: self.show_tour(DEFAULT_TOUR))

        self.tour_action = self.create_action(
            TourActions.ShowTour,
            text=_("Interactive tour"),
            icon=self.create_icon('tour'),
            triggered=lambda: self.show_tour(DEFAULT_TOUR)
        )

    # --- PluginMainContainer API
    # ------------------------------------------------------------------------
    def setup(self):
        self.tours_menu = self.create_menu(
            "tours_menu", _("Interactive tours"))

    def update_actions(self):
        pass

    # --- Public API
    # ------------------------------------------------------------------------
    def register_tour(self, tour_id, title, tour_data):
        """
        Register a new interactive tour on spyder.

        Parameters
        ----------
        tour_id: str
            Unique tour string identifier.
        title: str
            Localized tour name.
        tour_data: dict
            The tour steps.
        """
        if tour_id in self._tours:
            raise SpyderAPIError(
                "Tour with id '{}' has already been registered!".format(
                    tour_id))

        self._tours[tour_id] = tour_data
        self._tour_titles[tour_id] = title
        action = self.create_action(
            tour_id,
            text=title,
            triggered=lambda: self.show_tour(tour_id),
        )
        self.add_item_to_menu(action, menu=self.tours_menu)

    def show_tour(self, tour_id):
        """
        Show interactive tour.

        Parameters
        ----------
        tour_id: str
            Unique tour string identifier.
        """
        self.sig_unmaximize_plugin_requested.emit()
        tour_data = self._tours[tour_id]
        dic = {'last': 0, 'tour': tour_data}
        self._tour_widget.set_tour(tour_id, dic, self._main)
        self._tour_widget.start_tour()

    def show_tour_message(self):
        """
        Show message about starting the tour the first time Spyder starts.
        """
        self._tour_dialog.show()
        self._tour_dialog.raise_()



# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
# (see spyder/__init__.py for details)

"""Mouse shortcut editor dialog"""

# Standard library imports
from itertools import combinations
import sys

# Third party imports
from qtpy.QtCore import Qt, Signal
from qtpy.QtWidgets import (
    QCheckBox,
    QDialog,
    QDialogButtonBox,
    QHBoxLayout,
    QLabel,
    QSizePolicy,
    QVBoxLayout,
    QWidget,
)

# Local imports
from spyder.api.config.mixins import SpyderConfigurationAccessor
from spyder.api.translations import _
from spyder.api.widgets.dialogs import SpyderDialogButtonBox
from spyder.utils.icon_manager import ima
from spyder.widgets.helperwidgets import TipWidget


class MouseShortcutEditor(QDialog, SpyderConfigurationAccessor):
    """A dialog to edit the modifier keys for CodeEditor mouse interactions."""

    CONF_SECTION = "editor"

    def __init__(self, parent):
        super().__init__(parent)
        mouse_shortcuts = self.get_conf('mouse_shortcuts')
        self.setWindowFlags(
            self.windowFlags() & ~Qt.WindowContextHelpButtonHint
        )

        layout = QVBoxLayout()

        description = QLabel(
            "Here you can configure shortcuts that are triggered by one or "
            "more key modifiers and a left mouse click"
        )
        description.setWordWrap(True)
        layout.addWidget(description)

        self.scrollflag_shortcut = ShortcutSelector(
            self,
            _("Jump within the document in the scroll flags area"),
            mouse_shortcuts['jump_to_position']
        )
        self.scrollflag_shortcut.sig_changed.connect(self.validate)
        layout.addWidget(self.scrollflag_shortcut)

        self.goto_def_shortcut = ShortcutSelector(
            self,
            _("Go to a code definition"),
            mouse_shortcuts['goto_definition']
        )
        self.goto_def_shortcut.sig_changed.connect(self.validate)
        layout.addWidget(self.goto_def_shortcut)

        self.add_cursor_shortcut = ShortcutSelector(
            self,
            _("Add/remove an additional cursor"),
            mouse_shortcuts['add_remove_cursor']
        )
        self.add_cursor_shortcut.sig_changed.connect(self.validate)
        layout.addWidget(self.add_cursor_shortcut)

        self.column_cursor_shortcut = ShortcutSelector(
            self,
            _("Add a column of cursors"),
            mouse_shortcuts['column_cursor']
        )
        self.column_cursor_shortcut.sig_changed.connect(self.validate)
        layout.addWidget(self.column_cursor_shortcut)
        
        buttons = (
            QDialogButtonBox.StandardButton.Apply |
            QDialogButtonBox.StandardButton.Ok |
            QDialogButtonBox.StandardButton.Cancel
        )
        button_box = SpyderDialogButtonBox(buttons, parent=self)
        apply_b = button_box.button(QDialogButtonBox.StandardButton.Apply)
        apply_b.clicked.connect(self.apply_mouse_shortcuts)
        apply_b.setEnabled(False)
        self.apply_button = apply_b

        ok_b = button_box.button(QDialogButtonBox.StandardButton.Ok)
        ok_b.clicked.connect(self.accept)
        self.ok_button = ok_b

        cancel_b = button_box.button(QDialogButtonBox.StandardButton.Cancel)
        cancel_b.clicked.connect(self.reject)
        layout.addWidget(button_box)
        
        self.setLayout(layout)

    def apply_mouse_shortcuts(self):
        """Set new config to CONF"""
        self.set_conf('mouse_shortcuts', self.mouse_shortcuts)
        self.scrollflag_shortcut.apply_modifiers()
        self.goto_def_shortcut.apply_modifiers()
        self.add_cursor_shortcut.apply_modifiers()
        self.column_cursor_shortcut.apply_modifiers()
        self.apply_button.setEnabled(False)

    def accept(self):
        """Apply new settings and close dialog."""
        self.apply_mouse_shortcuts()
        super().accept()

    def validate(self):
        """
        Detect conflicts between shortcuts, and detect if current selection is
        different from current config. Set Ok and Apply buttons enabled or
        disabled accordingly, as well as set visibility of the warning for
        shortcut conflict.
        """
        shortcut_selectors = (
            self.scrollflag_shortcut,
            self.goto_def_shortcut,
            self.add_cursor_shortcut,
            self.column_cursor_shortcut
        )

        for selector in shortcut_selectors:
            selector.warning.setVisible(False)

        conflict = False
        for a, b in combinations(shortcut_selectors, 2):
            if a.modifiers() and a.modifiers() == b.modifiers():
                conflict = True
                a.warning.setVisible(True)
                b.warning.setVisible(True)

        self.ok_button.setEnabled(not conflict)

        self.apply_button.setEnabled(
            not conflict and (
                self.scrollflag_shortcut.is_changed() or
                self.goto_def_shortcut.is_changed() or
                self.add_cursor_shortcut.is_changed() or
                self.column_cursor_shortcut.is_changed()
            )
        )

    @property
    def mouse_shortcuts(self):
        """Format shortcuts dict for CONF."""
        return {
            'jump_to_position': self.scrollflag_shortcut.modifiers(),
            'goto_definition': self.goto_def_shortcut.modifiers(),
            'add_remove_cursor': self.add_cursor_shortcut.modifiers(),
            'column_cursor': self.column_cursor_shortcut.modifiers()
        }


class ShortcutSelector(QWidget):
    """Line representing an editor for a single mouse shortcut."""

    sig_changed = Signal()

    def __init__(self, parent, label, modifiers):
        super().__init__(parent)

        layout = QHBoxLayout()

        label = QLabel(label)
        layout.addWidget(label)

        spacer = QWidget(self)
        spacer.setSizePolicy(
            QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Preferred
        )
        layout.addWidget(spacer)

        self.ctrl_check = QCheckBox(
            "Cmd" if sys.platform == "darwin" else "Ctrl"
        )
        self.ctrl_check.setChecked("ctrl" in modifiers.lower())
        self.ctrl_check.toggled.connect(self.validate)
        layout.addWidget(self.ctrl_check)

        self.alt_check = QCheckBox("Alt")
        self.alt_check.setChecked("alt" in modifiers.lower())
        self.alt_check.toggled.connect(self.validate)
        layout.addWidget(self.alt_check)

        self.meta_check = QCheckBox("Meta")
        self.meta_check.setChecked("meta" in modifiers.lower())
        self.meta_check.toggled.connect(self.validate)
        layout.addWidget(self.meta_check)

        self.shift_check = QCheckBox("Shift")
        self.shift_check.setChecked("shift" in modifiers.lower())
        self.shift_check.toggled.connect(self.validate)
        layout.addWidget(self.shift_check)

        warning_icon = ima.icon("MessageBoxWarning")
        self.warning = TipWidget(
            _("Shortcut Conflicts With Another"),
            warning_icon,
            warning_icon
        )

        # Thanks to https://stackoverflow.com/a/34663079/3220135
        sp_retain = self.warning.sizePolicy()
        sp_retain.setRetainSizeWhenHidden(True)
        self.warning.setSizePolicy(sp_retain)
        self.warning.setVisible(False)
        layout.addWidget(self.warning)

        self.setLayout(layout)

        self.apply_modifiers()

    def validate(self):
        """
        Cannot have shortcut of Shift alone as that conflicts with setting the
        cursor position without moving the anchor. Enable/Disable the Shift
        checkbox accordingly. (Re)Emit a signal to MouseShortcutEditor which
        will perform other validation.
        """
        if (
            self.ctrl_check.isChecked() or
            self.alt_check.isChecked() or
            self.meta_check.isChecked()
        ):
            self.shift_check.setEnabled(True)
        else:
            self.shift_check.setEnabled(False)
            self.shift_check.setChecked(False)

        self.sig_changed.emit()

    def modifiers(self):
        """Get the current modifiers string."""
        modifiers = []
        if self.ctrl_check.isChecked():
            modifiers.append("Ctrl")
        if self.alt_check.isChecked():
            modifiers.append("Alt")
        if self.meta_check.isChecked():
            modifiers.append("Meta")
        if self.shift_check.isChecked():
            modifiers.append("Shift")
        return "+".join(modifiers)

    def is_changed(self):
        """Is the current selection different from when last applied?"""
        return self.current_modifiers != self.modifiers()

    def apply_modifiers(self):
        """Informs ShortcutSelector that settings have been applied."""
        self.current_modifiers = self.modifiers()



# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
# (see spyder/__init__.py for details)

"""
Profiler plugin API.
"""

# Local imports
from spyder.plugins.profiler.widgets.main_widget import (  # noqa
    ProfilerWidgetActions,
    ProfilerWidgetMenus,
    ProfilerContextMenuSections,
    ProfilerWidgetContextMenuActions,
    ProfilerWidgetMainToolbarSections,
)



# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
# (see spyder/__init__.py for details)

"""
Language Server Protocol advanced configuration tab.
"""

# Standard library imports
import re

# Third party imports
from qtpy.QtCore import Qt, Slot
from qtpy.QtWidgets import (QGroupBox, QGridLayout, QLabel, QMessageBox,
                            QVBoxLayout, QWidget)

# Local imports
from spyder.api.preferences import SpyderPreferencesTab
from spyder.api.translations import _
from spyder.utils.misc import check_connection_port


class AdvancedConfigTab(SpyderPreferencesTab):
    """PyLS advanced configuration tab."""

    TITLE = _('Advanced')

    def __init__(self, parent):
        super().__init__(parent)

        lsp_advanced_group = QGroupBox(_(
            'Python Language Server configuration'))
        advanced_label = QLabel(
            _("<b>Warning</b>: Only modify these values if "
              "you know what you're doing!"))
        advanced_label.setWordWrap(True)
        advanced_label.setAlignment(Qt.AlignJustify)

        # Advanced settings checkbox
        self.advanced_options_check = self.create_checkbox(
            _("Enable advanced settings"), 'advanced/enabled')

        # Advanced options
        self.advanced_module = self.create_lineedit(
            _("Module for the Python language server: "),
            'advanced/module', alignment=Qt.Horizontal,
            word_wrap=False)
        self.advanced_host = self.create_lineedit(
            _("IP Address and port to bind the server to: "),
            'advanced/host', alignment=Qt.Horizontal,
            word_wrap=False)
        self.advanced_port = self.create_spinbox(
            ":", "", 'advanced/port', min_=1, max_=65535, step=1)
        self.external_server = self.create_checkbox(
            _("This is an external server"),
            'advanced/external')
        self.use_stdio = self.create_checkbox(
            _("Use stdio pipes to communicate with server"),
            'advanced/stdio')
        self.use_stdio.checkbox.stateChanged.connect(self.disable_tcp)
        self.external_server.checkbox.stateChanged.connect(self.disable_stdio)

        # Advanced layout
        advanced_g_layout = QGridLayout()
        advanced_g_layout.addWidget(self.advanced_module.label, 1, 0)
        advanced_g_layout.addWidget(self.advanced_module.textbox, 1, 1)
        advanced_g_layout.addWidget(self.advanced_host.label, 2, 0)

        advanced_host_port_g_layout = QGridLayout()
        advanced_host_port_g_layout.addWidget(self.advanced_host.textbox, 1, 0)
        advanced_host_port_g_layout.addWidget(self.advanced_port.plabel, 1, 1)
        advanced_host_port_g_layout.addWidget(self.advanced_port.spinbox, 1, 2)
        advanced_g_layout.addLayout(advanced_host_port_g_layout, 2, 1)

        # External server and stdio options layout
        advanced_server_layout = QVBoxLayout()
        advanced_server_layout.addWidget(self.external_server)
        advanced_server_layout.addWidget(self.use_stdio)

        advanced_options_layout = QVBoxLayout()
        advanced_options_layout.addLayout(advanced_g_layout)
        advanced_options_layout.addLayout(advanced_server_layout)

        # Set advanced options enabled/disabled
        advanced_options_widget = QWidget()
        advanced_options_widget.setLayout(advanced_options_layout)
        advanced_options_widget.setEnabled(self.get_option('advanced/enabled'))
        self.advanced_options_check.checkbox.toggled.connect(
            advanced_options_widget.setEnabled)
        self.advanced_options_check.checkbox.toggled.connect(
            self.show_advanced_warning)

        # Advanced options layout
        advanced_layout = QVBoxLayout()
        advanced_layout.addWidget(advanced_label)
        advanced_layout.addWidget(self.advanced_options_check)
        advanced_layout.addWidget(advanced_options_widget)

        lsp_advanced_group.setLayout(advanced_layout)

        layout = QVBoxLayout()
        layout.addWidget(lsp_advanced_group)
        self.setLayout(layout)

    def disable_tcp(self, state):
        if state == Qt.Checked:
            self.advanced_host.textbox.setEnabled(False)
            self.advanced_port.spinbox.setEnabled(False)
            self.external_server.checkbox.stateChanged.disconnect()
            self.external_server.checkbox.setChecked(False)
            self.external_server.checkbox.setEnabled(False)
        else:
            self.advanced_host.textbox.setEnabled(True)
            self.advanced_port.spinbox.setEnabled(True)
            self.external_server.checkbox.setChecked(False)
            self.external_server.checkbox.setEnabled(True)
            self.external_server.checkbox.stateChanged.connect(
                self.disable_stdio)

    def disable_stdio(self, state):
        if state == Qt.Checked:
            self.advanced_host.textbox.setEnabled(True)
            self.advanced_port.spinbox.setEnabled(True)
            self.advanced_module.textbox.setEnabled(False)
            self.use_stdio.stateChanged.disconnect()
            self.use_stdio.setChecked(False)
            self.use_stdio.setEnabled(False)
        else:
            self.advanced_host.textbox.setEnabled(True)
            self.advanced_port.spinbox.setEnabled(True)
            self.advanced_module.textbox.setEnabled(True)
            self.use_stdio.setChecked(False)
            self.use_stdio.setEnabled(True)
            self.use_stdio.stateChanged.connect(self.disable_tcp)

    @Slot(bool)
    def show_advanced_warning(self, state):
        """
        Show a warning when trying to modify the PyLS advanced
        settings.
        """
        # Don't show warning if the option is already enabled.
        # This avoids showing it when the Preferences dialog
        # is created.
        if self.get_option('advanced/enabled'):
            return

        # Show warning when toggling the button state
        if state:
            QMessageBox.warning(
                self,
                _("Warning"),
                _("<b>Modifying these options can break code completion!!</b>"
                  "<br><br>"
                  "If that's the case, please reset your Spyder preferences "
                  "by going to the menu"
                  "<br><br>"
                  "<tt>Tools > Reset all preferences to default</tt>"
                  "<br><br>"
                  "instead of reporting a bug."))

    def is_valid(self):
        host = self.advanced_host.textbox.text()

        # If host is not local, the server must be external
        # and we need to automatically check the corresponding
        # option
        if host not in ['127.0.0.1', 'localhost']:
            self.external_server.checkbox.setChecked(True)

        # Checks for external PyLS
        if self.external_server.checkbox.isChecked():
            port = int(self.advanced_port.spinbox.text())

            # Check that host and port of the current server are
            # different from the new ones provided to connect to
            # an external server.
            lsp = self.plugin.get_provider('lsp')
            pyclient = lsp.clients.get('python')
            if pyclient is not None:
                instance = pyclient['instance']
                if (instance is not None and
                        not pyclient['config']['external']):
                    if (instance.server_host == host and
                            instance.server_port == port):
                        self.report_no_address_change()
                        return False

            # Check connection to LSP server using a TCP socket
            response = check_connection_port(host, port)
            if not response:
                self.report_no_external_server(host, port, 'python')
                return False

        return True

    def report_no_external_server(self, host, port, language):
        """
        Report that connection couldn't be established with
        an external server.
        """
        QMessageBox.critical(
            self,
            _("Error"),
            _("It appears there is no {language} language server listening "
              "at address:"
              "<br><br>"
              "<tt>{host}:{port}</tt>"
              "<br><br>"
              "Please verify that the provided information is correct "
              "and try again.").format(host=host, port=port,
                                       language=language.capitalize())
        )

    def report_no_address_change(self):
        """
        Report that server address has no changed after checking the
        external server option.
        """
        QMessageBox.critical(
            self,
            _("Error"),
            _("The address of the external server you are trying to connect "
              "to is the same as the one of the current internal server "
              "started by Spyder."
              "<br><br>"
              "Please provide a different address!")
        )



# -----------------------------------------------------------------------------
# Copyright (c) 2021- Spyder Project Contributors
#
# Released under the terms of the MIT License
# (see LICENSE.txt in the project root directory for details)
# -----------------------------------------------------------------------------

"""Plugin registry configuration page."""

# Third party imports
from pyuca import Collator
from qtpy.QtWidgets import QVBoxLayout, QLabel

# Local imports
from spyder.api.preferences import PluginConfigPage
from spyder.api.translations import _
from spyder.utils.palette import SpyderPalette
from spyder.widgets.elementstable import ElementsTable
from spyder.widgets.helperwidgets import FinderWidget


class PluginsConfigPage(PluginConfigPage):

    def setup_page(self):
        newcb = self.create_checkbox
        self.plugins_checkboxes = {}

        header_label = QLabel(
            _(
                "Disable a Spyder plugin (external or built-in) to prevent it "
                "from loading until re-enabled here, to simplify the interface "
                "or in case it causes problems."
            )
        )
        header_label.setWordWrap(True)

        # To save the plugin elements
        internal_elements = []
        external_elements = []

        # ------------------ Internal plugins ---------------------------------
        for plugin_name in self.plugin.all_internal_plugins:
            (conf_section_name, PluginClass) = (
                self.plugin.all_internal_plugins[plugin_name]
            )

            if not getattr(PluginClass, "CAN_BE_DISABLED", True):
                # Do not list core plugins that can not be disabled
                continue

            plugin_state = self.get_option(
                "enable", section=conf_section_name, default=True
            )
            cb = newcb(
                "",
                "enable",
                default=True,
                section=conf_section_name,
                restart=True,
            )

            internal_elements.append(
                dict(
                    title=PluginClass.get_name(),
                    description=PluginClass.get_description(),
                    icon=PluginClass.get_icon(),
                    widget=cb,
                    additional_info=_("Built-in"),
                    additional_info_color=SpyderPalette.COLOR_TEXT_4,
                )
            )

            self.plugins_checkboxes[plugin_name] = (cb.checkbox, plugin_state)

        # ------------------ External plugins ---------------------------------
        for plugin_name in self.plugin.all_external_plugins:
            (conf_section_name, PluginClass) = (
                self.plugin.all_external_plugins[plugin_name]
            )

            if not getattr(PluginClass, "CAN_BE_DISABLED", True):
                # Do not list external plugins that can not be disabled
                continue

            plugin_state = self.get_option(
                f"{conf_section_name}/enable",
                section=self.plugin._external_plugins_conf_section,
                default=True,
            )
            cb = newcb(
                "",
                f"{conf_section_name}/enable",
                default=True,
                section=self.plugin._external_plugins_conf_section,
                restart=True,
            )

            external_elements.append(
                dict(
                    title=PluginClass.get_name(),
                    description=PluginClass.get_description(),
                    icon=PluginClass.get_icon(),
                    widget=cb,
                )
            )

            self.plugins_checkboxes[plugin_name] = (cb.checkbox, plugin_state)

        # Sort elements by title for easy searching
        collator = Collator()
        internal_elements.sort(key=lambda e: collator.sort_key(e["title"]))
        external_elements.sort(key=lambda e: collator.sort_key(e["title"]))

        # Build plugins table, showing external plugins first.
        self._plugins_table = ElementsTable(
            self, add_padding_around_widgets=True
        )
        self._plugins_table.setup_elements(
            external_elements + internal_elements
        )

        # Finder to filter plugins
        finder = FinderWidget(
            self,
            find_on_change=True,
            show_close_button=False,
            set_min_width=False,
        )
        finder.sig_find_text.connect(self._do_find)

        # Layout
        layout = QVBoxLayout()
        layout.addWidget(header_label)
        layout.addSpacing(15)
        layout.addWidget(self._plugins_table)
        layout.addWidget(finder)
        layout.addSpacing(15)
        self.setLayout(layout)

    def apply_settings(self):
        for plugin_name in self.plugins_checkboxes:
            cb, previous_state = self.plugins_checkboxes[plugin_name]
            if cb.isChecked() and not previous_state:
                self.plugin.set_plugin_enabled(plugin_name)
                PluginClass = None
                external = False
                if plugin_name in self.plugin.all_internal_plugins:
                    (__, PluginClass) = self.plugin.all_internal_plugins[
                        plugin_name
                    ]
                elif plugin_name in self.plugin.all_external_plugins:
                    (__, PluginClass) = self.plugin.all_external_plugins[
                        plugin_name
                    ]
                    external = True  # noqa

                # TODO: Once we can test that all plugins can be restarted
                # without problems during runtime, we can enable the
                # autorestart feature provided by the plugin registry:
                # self.plugin.register_plugin(self.main, PluginClass,
                #                             external=external)
            elif not cb.isChecked() and previous_state:
                # TODO: Once we can test that all plugins can be restarted
                # without problems during runtime, we can enable the
                # autorestart feature provided by the plugin registry:
                # self.plugin.delete_plugin(plugin_name)
                pass
        return set({})

    def _do_find(self, text):
        self._plugins_table.do_find(text)



# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
# (see spyder/__init__.py for details)

"""Dialog to handle remote connections."""

# Standard library imports
from __future__ import annotations

# Third party imports
from qtpy.QtCore import Signal
from qtpy.QtWidgets import (
    QDialogButtonBox,
    QHBoxLayout,
    QMessageBox,
    QPushButton,
)

# Local imports
from spyder.api.translations import _
from spyder.api.widgets.dialogs import SpyderDialogButtonBox
from spyder.plugins.remoteclient.api.protocol import (
    ConnectionInfo,
    ConnectionStatus,
    ClientType,
)
from spyder.plugins.remoteclient.widgets.connectionpages import (
    ConnectionPage,
    CreateEnvMethods,
    ENV_MANAGER,
    NewConnectionPage,
)
from spyder.utils.icon_manager import ima
from spyder.utils.stylesheet import MAC, WIN
from spyder.widgets.sidebardialog import SidebarDialog


class ConnectionDialog(SidebarDialog):
    """
    Dialog to handle and display remote connection information for different
    machines.
    """

    TITLE = _("Remote connections")
    FIXED_SIZE = True
    MIN_WIDTH = 895 if MAC else (810 if WIN else 860)
    MIN_HEIGHT = 740 if MAC else (655 if WIN else 690)
    PAGE_CLASSES = [NewConnectionPage]

    sig_start_server_requested = Signal(str)
    sig_stop_server_requested = Signal(str)
    sig_abort_connection_requested = Signal(str)
    sig_server_renamed = Signal(str)
    sig_connections_changed = Signal()
    sig_server_updated = Signal(str)
    sig_create_env_requested = Signal(str, str, str, list)
    sig_import_env_requested = Signal(str, str, str)

    def __init__(self, parent=None):
        self.ICON = ima.icon('remote_server')
        super().__init__(parent)
        self._container = parent

        # -- Setup
        self._add_saved_connection_pages()

        # If there's more than one page, give focus to the first server because
        # users will probably want to interact with servers here rather than
        # create new connections.
        if self.number_of_pages() > 1:
            # Index 1 is the separator added after the new connection page
            self.set_current_index(2)

        # -- Signals
        if self._container is not None:
            self._container.sig_connection_status_changed.connect(
                self._on_connection_status_changed
            )

        if ENV_MANAGER:
            self._new_connection_page: NewConnectionPage = self.get_page(0)
            self._new_connection_page.tabs.currentChanged.connect(
                self._on_new_connection_page_tab_changed
            )
            self._new_connection_page.env_method_group.idToggled.connect(
                self._set_buttons_for_env_creation_method
            )

    # ---- SidebarDialog API
    # -------------------------------------------------------------------------
    def create_buttons(self):
        bbox = SpyderDialogButtonBox(QDialogButtonBox.Cancel)

        self._button_cancel = bbox.button(QDialogButtonBox.Cancel)

        self._button_save_connection = QPushButton(_("Save"))
        self._button_save_connection.clicked.connect(
            self._save_connection_info
        )
        bbox.addButton(
            self._button_save_connection, QDialogButtonBox.ResetRole
        )

        self._button_remove_connection = QPushButton(_("Remove"))
        self._button_remove_connection.clicked.connect(
            self._remove_connection_info
        )
        bbox.addButton(
            self._button_remove_connection, QDialogButtonBox.ResetRole
        )

        self._button_clear_settings = QPushButton(_("Clear"))
        self._button_clear_settings.clicked.connect(self._clear_settings)
        bbox.addButton(
            self._button_clear_settings, QDialogButtonBox.ActionRole
        )

        self._button_stop = QPushButton(_("Stop"))
        self._button_stop.clicked.connect(self._on_button_stop_clicked)
        bbox.addButton(self._button_stop, QDialogButtonBox.ActionRole)

        self._button_connect = QPushButton(_("Connect"))
        self._button_connect.clicked.connect(self._start_server)
        bbox.addButton(self._button_connect, QDialogButtonBox.ActionRole)

        self._button_next = QPushButton(_("Next"))
        self._button_next.clicked.connect(self._on_button_next_clicked)
        bbox.addButton(self._button_next, QDialogButtonBox.ActionRole)

        self._button_back = QPushButton(_("Back"))
        self._button_back.clicked.connect(self._on_back_button_clicked)
        bbox.addButton(self._button_back, QDialogButtonBox.ResetRole)

        layout = QHBoxLayout()
        layout.addWidget(bbox)

        return bbox, layout

    def current_page_changed(self, index):
        """Update the state of buttons when moving from page to page."""
        page = self.get_page(index)
        if page.NEW_CONNECTION:
            self._button_save_connection.setEnabled(True)
            self._button_clear_settings.setHidden(False)
            self._button_remove_connection.setHidden(True)
            self._button_stop.setHidden(True)
            self._button_cancel.setText("Cancel")

            if ENV_MANAGER:
                if page.get_current_tab() == "SSH":
                    if page.is_ssh_info_widget_shown():
                        self._button_connect.setHidden(True)
                        self._button_next.setHidden(False)
                        self._button_back.setHidden(True)
                        self._button_save_connection.setEnabled(False)
                    elif page.is_env_creation_widget_shown():
                        self._set_buttons_for_env_creation_method()
                    else:
                        self._set_buttons_for_env_packages_list()
                else:
                    self._button_connect.setHidden(False)
                    self._button_next.setHidden(True)
                    self._button_back.setHidden(True)
                    self._button_save_connection.setEnabled(True)
            else:
                self._button_connect.setHidden(False)
                self._button_next.setHidden(True)
                self._button_back.setHidden(True)
        else:
            # If the page info is not changed, then users only need to close
            # the dialog
            if page.is_modified:
                self._button_save_connection.setEnabled(True)
                self._button_cancel.setText("Cancel")
            else:
                self._button_save_connection.setEnabled(False)
                self._button_cancel.setText("Close")

            # It's possible to remove the connection for saved pages
            self._button_remove_connection.setHidden(False)

            # These buttons are only needed in the New connection page
            for button in [
                self._button_clear_settings,
                self._button_next,
                self._button_back,
            ]:
                button.setHidden(True)

            # The state of the other buttons can be updated according to the
            # connection status
            self._update_buttons_state_per_status(page.status)

    # ---- Private API
    # -------------------------------------------------------------------------
    def _save_connection_info(self):
        """Save the connection info stored in a page."""
        page = self.get_page()

        # Validate info
        if not page.validate_page():
            return

        if page.NEW_CONNECTION:
            # Save info provided by users
            page.save_to_conf()

            # Add separator if needed
            if self.number_of_pages() == 1:
                self.add_separator()

            # Add connection page to the dialog with the new info
            self._add_connection_page(host_id=page.host_id, new=True)

            # Give focus to the new page
            self.set_current_index(self.number_of_pages() - 1)

            # Reset page in case users want to introduce another connection
            page.reset_page()
        else:
            # Update name in the dialog if it was changed by users. This needs
            # to be done before calling save_to_conf so that we can compare the
            # saved name with the current one.
            if page.has_new_name():
                self.get_item().setText(page.new_name)

            # Update connection info
            page.save_to_conf()

            # After saving to our config system, we can inform the container
            # that a change in connections took place.
            if page.new_name is not None:
                self.sig_server_renamed.emit(page.host_id)
                page.new_name = None

            # Mark page as not modified and disable save button
            page.set_modified(False)
            self._button_save_connection.setEnabled(False)

            # Update connection info if necessary
            page.update_connection_info()
            self.sig_server_updated.emit(page.host_id)

        # Inform container that a change in connections took place
        self.sig_connections_changed.emit()

    def _remove_connection_info(self):
        """
        Remove the connection info stored in a given page and hide it as well.
        """
        page = self.get_page()
        if not page.NEW_CONNECTION:
            reply = QMessageBox.question(
                self,
                _("Remove connection"),
                _(
                    "Do you want to remove the connection called <b>{}</b>?"
                ).format(page.get_name()),
                QMessageBox.Yes,
                QMessageBox.No,
            )

            if reply == QMessageBox.Yes:
                self.hide_page()
                page.remove_config_options()

                # Inform container that a change in connections took place
                self.sig_connections_changed.emit()

    def _clear_settings(self):
        """Clear the setting introduced in the new connection page."""
        page = self.get_page()
        if page.NEW_CONNECTION:
            page.reset_page(clear=True)

    def _start_server(self):
        """Start the server corresponding to a given page."""
        page = self.get_page()

        # Validate connection info.
        # NOTE: It always needs to be done in case the info changed (e.g. key
        # file could have been moved) before establishing the connection.
        if not page.validate_page():
            return

        # Validate env info for SSH connections
        if (
            ENV_MANAGER
            and page.NEW_CONNECTION
            and page.get_client_type() == ClientType.SSH
        ):
            if (
                self._new_connection_page.is_env_creation_widget_shown()
                and not self._new_connection_page.validate_env_creation()
            ):
                return
            elif (
                self._new_connection_page.is_env_packages_widget_shown()
                and not self._new_connection_page.get_env_packages_list()
            ):
                return

        # This uses the current host_id in case users want to start a
        # connection directly from the new connection page (
        # _save_connection_info generates a new id fo that page at the end).
        host_id = page.host_id

        if page.NEW_CONNECTION or page.is_modified:
            # Save connection info if necessary
            self._save_connection_info()

        # Create env if requested for SSH connections
        if (
            ENV_MANAGER
            and page.NEW_CONNECTION
            and page.get_client_type() == ClientType.SSH
        ):
            if page.selected_env_creation_method() == CreateEnvMethods.NewEnv:
                env_name, python_version = page.get_create_env_info()
                packages_list = page.get_env_packages_list()
                self.sig_create_env_requested.emit(
                    host_id, env_name, python_version, packages_list
                )
            elif (
                page.selected_env_creation_method()
                == CreateEnvMethods.ImportEnv
            ):
                import_file_path, env_name = page.get_create_env_info()
                self.sig_import_env_requested.emit(
                    host_id, import_file_path, env_name
                )
            elif page.selected_env_creation_method() == CreateEnvMethods.NoEnv:
                self.sig_start_server_requested.emit(host_id)

            # Show again info widget in case users want to enter another
            # connection with similar settings.
            page.show_ssh_info_widget()
        else:
            self.sig_start_server_requested.emit(host_id)

    def _stop_server(self):
        """Stop the server corresponding to a given page."""
        page = self.get_page()

        # The stop button is not visible in the new connection page
        if not page.NEW_CONNECTION:
            self._button_stop.setHidden(True)
            self.sig_stop_server_requested.emit(page.host_id)

    def _add_connection_page(self, host_id: str, new: bool):
        """Add a new connection page to the dialog."""
        page = ConnectionPage(self, host_id=host_id)

        # This is necessary to make button_save_connection enabled when there
        # are config changes in the page
        page.apply_button_enabled.connect(
            self._update_buttons_state_on_info_change
        )

        if new:
            page.save_server_id()

        self.add_page(page)

        # Add saved logs to the page
        if self._container is not None:
            page.add_logs(self._container.client_logs.get(host_id, []))

            # This updates the info shown in the "Connection info" tab of pages
            self._container.sig_connection_status_changed.connect(
                page.update_status
            )
            self._container.sig_client_message_logged.connect(page.add_log)

    def _add_saved_connection_pages(self):
        """Add a connection page for each server saved in our config system."""
        page = self.get_page(index=0)
        servers = page.get_option("servers", default={})

        if servers:
            self.add_separator()

            for id_ in servers.keys():
                self._add_connection_page(host_id=id_, new=False)

    def _update_buttons_state_on_info_change(self, state: bool):
        """Update the state of buttons when connection info changes."""
        self._button_save_connection.setEnabled(state)
        self._button_cancel.setText("Cancel")

    def _update_buttons_state_per_status(self, status: ConnectionStatus):
        """Update buttons state according to the connection status."""
        if status in [ConnectionStatus.Inactive, ConnectionStatus.Error]:
            self._button_connect.setHidden(False)
        else:
            self._button_connect.setHidden(True)

        if status in [ConnectionStatus.Active, ConnectionStatus.Connecting]:
            self._button_stop.setHidden(False)
            self._button_stop.setEnabled(True)
        elif status in [
            ConnectionStatus.Inactive,
            ConnectionStatus.Error,
            ConnectionStatus.Stopping,
        ]:
            self._button_stop.setHidden(True)
        else:
            self._button_stop.setHidden(False)
            self._button_stop.setEnabled(False)

        if status in [
            ConnectionStatus.Active,
            ConnectionStatus.Connecting,
            ConnectionStatus.Connected,
            ConnectionStatus.Starting,
            ConnectionStatus.Stopping,
        ]:
            self._button_remove_connection.setEnabled(False)
        else:
            self._button_remove_connection.setEnabled(True)

    def _set_buttons_for_env_creation_method(
        self, id_: CreateEnvMethods | None = None
    ):
        if id_ is None:
            id_ = self._new_connection_page.selected_env_creation_method()

        # When creating a new env, users need to provide a list of packages for
        # it, so the connection can't be established yet
        if id_ == CreateEnvMethods.NewEnv:
            self._button_connect.setHidden(True)
            self._button_next.setHidden(False)
        else:
            self._button_connect.setHidden(False)
            self._button_next.setHidden(True)

        # Connection info can be saved if users decide to create no env
        if id_ == CreateEnvMethods.NoEnv:
            self._button_save_connection.setEnabled(True)
        else:
            self._button_save_connection.setEnabled(False)

        # The back button will always be visible in this case.
        self._button_back.setHidden(False)

    def _set_buttons_for_env_packages_list(self):
        # We can create a connection at this point
        self._button_connect.setHidden(False)

        # There are no additional subpages to go with Next
        self._button_next.setHidden(True)

        # We can't save the connection info if users are selecting packages for
        # their remote env.
        self._button_save_connection.setEnabled(False)

        # The back button will always be visible in this case.
        self._button_back.setHidden(False)

    def _on_button_next_clicked(self):
        page = self._new_connection_page

        if page.is_ssh_info_widget_shown():
            # Validate info
            if not page.validate_page():
                return

            page.show_env_creation_widget()
            self._set_buttons_for_env_creation_method()
        else:
            # Validate env creation info
            if not page.validate_env_creation():
                return

            page.setup_env_packages_widget()
            page.show_env_packages_widget()
            self._set_buttons_for_env_packages_list()

    def _on_back_button_clicked(self):
        page = self._new_connection_page

        if page.is_env_packages_widget_shown():
            page.show_env_creation_widget()
            self._set_buttons_for_env_creation_method()
        else:
            page.show_ssh_info_widget()
            self._button_back.setHidden(True)
            self._button_connect.setHidden(True)
            self._button_next.setHidden(False)

        self._button_save_connection.setEnabled(False)

    def _on_button_stop_clicked(self):
        page = self.get_page()
        if page.NEW_CONNECTION:
            return

        if page.status == ConnectionStatus.Active:
            self._stop_server()
        elif page.status == ConnectionStatus.Connecting:
            self.sig_abort_connection_requested.emit(page.host_id)

    def _on_new_connection_page_tab_changed(self, index):
        page = self._new_connection_page
        if page.get_current_tab(index) == "SSH":
            if page.is_ssh_info_widget_shown():
                self._button_connect.setHidden(True)
                self._button_next.setHidden(False)
                self._button_back.setHidden(True)
                self._button_save_connection.setEnabled(False)
            elif page.is_env_creation_widget_shown():
                self._set_buttons_for_env_creation_method()
            else:
                self._set_buttons_for_env_packages_list()
        else:
            self._button_connect.setHidden(False)
            self._button_next.setHidden(True)
            self._button_back.setHidden(True)
            self._button_save_connection.setEnabled(True)

    def _on_connection_status_changed(self, info: ConnectionInfo):
        page = self.get_page()
        if page.host_id == info["id"]:
            self._update_buttons_state_per_status(info["status"])


def test():
    from spyder.utils.qthelpers import qapplication
    from spyder.utils.stylesheet import APP_STYLESHEET

    app = qapplication()  # noqa
    app.setStyleSheet(str(APP_STYLESHEET))

    dialog = ConnectionDialog()
    dialog.exec_()


if __name__ == "__main__":
    test()



# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
# (see spyder/__init__.py for details)

"""Module serving the "About Spyder" function"""

# Standard library imports
import sys

# Third party imports
import qstylizer.style
from qtpy.QtCore import Qt
from qtpy.QtWidgets import (QApplication, QDialog, QDialogButtonBox,
                            QHBoxLayout, QVBoxLayout, QLabel, QLayout,
                            QPushButton, QScrollArea, QTabWidget)

# Local imports
from spyder import (
    __project_url__ as project_url,
    __forum_url__ as forum_url,
    __trouble_url__ as trouble_url,
    __website_url__ as website_url,
    get_versions, get_versions_text
)
from spyder.api.translations import _
from spyder.api.widgets.dialogs import SpyderDialogButtonBox
from spyder.api.widgets.mixins import SvgToScaledPixmap
from spyder.utils.icon_manager import ima
from spyder.utils.stylesheet import (
    AppStyle,
    DialogStyle,
    MAC,
    PREFERENCES_TABBAR_STYLESHEET,
    WIN
)


class AboutDialog(QDialog, SvgToScaledPixmap):

    PADDING = 5 if MAC else 15

    def __init__(self, parent):
        """Create About Spyder dialog with general information."""
        QDialog.__init__(self, parent)

        self.setWindowFlags(
            self.windowFlags() & ~Qt.WindowContextHelpButtonHint
        )
        self.setWindowTitle(_("About Spyder"))
        self.setWindowIcon(ima.icon("MessageBoxInformation"))
        versions = get_versions()

        # -- Show Git revision for development version
        revlink = ''
        if versions['revision']:
            rev = versions['revision']
            revlink = ("<a href='https://github.com/spyder-ide/spyder/"
                       "commit/%s'>%s</a>" % (rev, rev))

        # -- Style attributes
        font_family = self.font().family()
        font_size = DialogStyle.ContentFontSize

        # -- Labels
        twitter_url = "https://twitter.com/Spyder_IDE",
        facebook_url = "https://www.facebook.com/SpyderIDE",
        youtube_url = "https://www.youtube.com/Spyder-IDE",
        instagram_url = "https://www.instagram.com/spyderide/",
        self.label_overview = QLabel(
            f"""
            <style>
                p, h1 {{margin-bottom: 2em}}
                h1 {{margin-top: 0}}
            </style>

            <div style='font-family: "{font_family}";
                        font-size: {font_size};
                        font-weight: normal;
                        '>
            <br>
            <h1>Spyder IDE</h1>

            <p>
            The Scientific Python Development Environment
            <br>
            <a href="{website_url}">Spyder-IDE.org</a>
            </p>

            <p>
            Python {versions['python']} {versions['bitness']}-bit |
            Qt {versions['qt']} |
            {versions['qt_api']} {versions['qt_api_ver']}
            <br>
            {versions['system']} {versions['release']} ({versions['machine']})
            </p>

            <p>
            <a href="{project_url}">GitHub</a> | <a href="{twitter_url}">
            Twitter</a> |
            <a href="{facebook_url}">Facebook</a> | <a href="{youtube_url}">
            YouTube</a> |
            <a href="{instagram_url}">Instagram</a>
            </p>

            </div>"""
        )

        self.label_community = QLabel(
            f"""
            <div style='font-family: "{font_family}";
                        font-size: {font_size};
                        font-weight: normal;
                        '>
            <br>
            Created by Pierre Raybaut; current maintainer is Carlos Cordoba.
            Developed by the
            <a href="{project_url}/graphs/contributors">international
            Spyder community</a>. Many thanks to all the Spyder beta testers
            and dedicated users.
            <p>For help with Spyder errors and crashes, please read our
            <a href="{trouble_url}">Troubleshooting Guide</a>, and for bug
            reports and feature requests, visit our
            <a href="{project_url}">Github site</a>. For project discussion,
            see our <a href="{forum_url}">Google Group</a>.
            </p>
            <p>
            This project is part of a larger effort to promote and
            facilitate the use of Python for scientific and engineering
            software development.
            The popular Python distributions
            <a href="https://www.anaconda.com/download/">Anaconda</a> and
            <a href="https://winpython.github.io/">WinPython</a>
            also contribute to this plan.
            </p>
            </div>""")

        self.label_legal = QLabel(
            f"""
            <div style='font-family: "{font_family}";
                        font-size: {font_size};
                        font-weight: normal;
                        '>
            <br>
            Copyright &copy; 2009-2020 Spyder Project Contributors and
            <a href="{project_url}/blob/master/AUTHORS.txt">others</a>.
            Distributed under the terms of the
            <a href="{project_url}/blob/master/LICENSE.txt">MIT License</a>.
            <p>
            <p>Certain source files under other compatible permissive
            licenses and/or originally by other authors.
            Spyder 3 theme icons derived from
            <a href="https://fontawesome.com/">Font Awesome</a> 4.7
            (&copy; 2016 David Gandy; SIL OFL 1.1) and
            <a href="http://materialdesignicons.com/">Material Design</a>
            (&copy; 2014 Austin Andrews; SIL OFL 1.1).
            </p>
            <p>
            Splash screen photo by
            <a href="https://unsplash.com/@benchaccounting?utm_source=
            unsplash&utm_medium=referral&utm_content=creditCopyText">Bench
            Accounting</a> on <a href="https://unsplash.com/?utm_source=
            unsplash&utm_medium=referral&utm_content=creditCopyText">Unsplash
            </a>.
            </p>
            <p>
            See the
            <a href="{project_url}/blob/master/NOTICE.txt">NOTICE</a>
            file for full legal information.
            </p>
            </div>
            """)

        for label in [self.label_overview, self.label_community,
                      self.label_legal]:
            label.setWordWrap(True)
            label.setAlignment(Qt.AlignTop)
            label.setOpenExternalLinks(True)
            label.setTextInteractionFlags(Qt.TextBrowserInteraction)
            label.setContentsMargins(
                (3 if MAC else 1) * self.PADDING,
                0,
                (3 if MAC else 1) * self.PADDING,
                (3 if MAC else 1) * self.PADDING,
            )

        self.label_pic = QLabel(self)
        self.label_pic.setPixmap(
            self.svg_to_scaled_pixmap("spyder_about", rescale=0.45)
        )
        self.label_pic.setAlignment(Qt.AlignBottom)

        self.info = QLabel(
            f"""
            <div style='font-family: "{font_family}";
                font-size: {font_size};
                font-weight: normal;
                '>
            {versions['spyder']}
            <br>{revlink}
            <br>({versions['installer']})
            <br>
            """
        )
        self.info.setAlignment(Qt.AlignHCenter)

        # -- Scroll areas
        scroll_overview = QScrollArea(self)
        scroll_overview.setWidgetResizable(True)
        scroll_overview.setWidget(self.label_overview)

        scroll_community = QScrollArea(self)
        scroll_community.setWidgetResizable(True)
        scroll_community.setWidget(self.label_community)

        scroll_legal = QScrollArea(self)
        scroll_legal.setWidgetResizable(True)
        scroll_legal.setWidget(self.label_legal)

        # Style for scroll areas needs to be applied after creating them.
        # Otherwise it doesn't have effect.
        for scroll_area in [scroll_overview, scroll_community, scroll_legal]:
            scroll_area.setStyleSheet(self._scrollarea_stylesheet)

        # -- Tabs
        self.tabs = QTabWidget(self)
        self.tabs.addTab(scroll_overview, _('Overview'))
        self.tabs.addTab(scroll_community, _('Community'))
        self.tabs.addTab(scroll_legal, _('Legal'))
        self.tabs.setElideMode(Qt.ElideNone)
        self.tabs.setStyleSheet(self._tabs_stylesheet)

        # -- Buttons
        bbox = SpyderDialogButtonBox(QDialogButtonBox.Ok)
        info_btn = QPushButton(_("Copy version info"))
        bbox.addButton(info_btn, QDialogButtonBox.ActionRole)

        # Apply style to buttons
        bbox.setStyleSheet(self._button_stylesheet)

        # -- Widget setup
        self.setWindowIcon(ima.icon('MessageBoxInformation'))
        self.setModal(False)

        # -- Layout
        piclayout = QVBoxLayout()
        piclayout.addStretch()
        piclayout.addWidget(self.label_pic)
        piclayout.addSpacing(-5)
        piclayout.addWidget(self.info)
        piclayout.addStretch()
        piclayout.setContentsMargins(
            # This makes the left and right margins around the image and info
            # to be the same on Linux and Windows.
            self.PADDING - (0 if MAC else 1) * AppStyle.MarginSize,
            0,
            self.PADDING,
            0
        )

        tabslayout = QHBoxLayout()
        tabslayout.addWidget(self.tabs)
        tabslayout.setSizeConstraint(QLayout.SizeConstraint.SetFixedSize)
        tabslayout.setContentsMargins(0, self.PADDING, 0, 0)

        btmhlayout = QHBoxLayout()
        btmhlayout.addStretch(1)
        btmhlayout.addWidget(bbox)
        btmhlayout.setContentsMargins(0, 0, self.PADDING, self.PADDING)
        btmhlayout.addStretch()

        vlayout = QVBoxLayout()
        vlayout.addLayout(tabslayout)
        vlayout.addLayout(btmhlayout)
        vlayout.setSizeConstraint(QLayout.SizeConstraint.SetFixedSize)

        mainlayout = QHBoxLayout(self)
        mainlayout.addLayout(piclayout)
        # This compensates the margin set for scroll areas to center them on
        # the tabbar
        mainlayout.addSpacing(-self.PADDING)
        mainlayout.addLayout(vlayout)

        # -- Signals
        info_btn.clicked.connect(self.copy_to_clipboard)
        bbox.accepted.connect(self.accept)

        # -- Style
        size = (600, 460) if MAC else ((585, 450) if WIN else (610, 455))
        self.setFixedSize(*size)
        self.setStyleSheet(self._main_stylesheet)

    def copy_to_clipboard(self):
        QApplication.clipboard().setText(get_versions_text())

    @property
    def _main_stylesheet(self):
        tabs_stylesheet = PREFERENCES_TABBAR_STYLESHEET.get_copy()
        css = tabs_stylesheet.get_stylesheet()

        # Set background color
        for widget in ["QDialog", "QLabel"]:
            css[widget].setValues(
                backgroundColor=DialogStyle.BackgroundColor
            )

        return css.toString()

    @property
    def _scrollarea_stylesheet(self):
        css = qstylizer.style.StyleSheet()

        # This is the only way to make the scroll areas to have the same
        # background color as the other widgets in the dialog.
        css.setValues(
            backgroundColor=DialogStyle.BackgroundColor
        )

        css.QScrollArea.setValues(
            # Default border color doesn't have enough contrast with the
            # background.
            border=f"1px solid {DialogStyle.BorderColor}",
            # This is necessary to center the tabbar on the scroll area
            marginLeft=f"{self.PADDING}px"
        )

        css.QScrollBar.setValues(
            # Default border color doesn't have enough contrast with the
            # background.
            border=f"1px solid {DialogStyle.BorderColor}",
        )

        return css.toString()

    @property
    def _button_stylesheet(self):
        css = qstylizer.style.StyleSheet()

        # Increase font size and padding
        css.QPushButton.setValues(
            fontSize=DialogStyle.ButtonsFontSize,
            padding=DialogStyle.ButtonsPadding
        )

        return css.toString()

    @property
    def _tabs_stylesheet(self):
        css = qstylizer.style.StyleSheet()

        # This fixes a visual glitch with the tabbar background color
        css.setValues(
            backgroundColor=DialogStyle.BackgroundColor
        )

        css['QTabWidget::pane'].setValues(
            # Set tab pane margins according to the dialog contents and layout
            marginTop=f"{(3 if MAC else 2) * AppStyle.MarginSize}px",
            marginRight=f"{self.PADDING}px",
            marginBottom=f"{(0 if MAC else 2) * AppStyle.MarginSize}px",
            marginLeft="0px",
            # Padding is not necessary in this case because we set a border for
            # the scroll areas.
            padding="0px",
        )

        return css.toString()


def test():
    """Run about widget test"""

    from spyder.utils.qthelpers import qapplication
    app = qapplication()  # noqa
    abt = AboutDialog(None)
    abt.show()
    sys.exit(abt.exec_())


if __name__ == '__main__':
    test()



# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
# (see spyder/__init__.py for details)

"""
Source code analysis utilities.
"""

import re

# Local import
from spyder.config.base import get_debug_level

DEBUG_EDITOR = get_debug_level() >= 3

# =============================================================================
# Find tasks - TODOs
# =============================================================================
TASKS_PATTERN = (
    r"#\s*(TODO|todo|FIXME|fixme|XXX|xxx|HINT|hint|TIP|tip|@todo|@TODO|"
    r"HACK|hack|BUG|bug|OPTIMIZE|optimize|!!!|\?\?\?)([^#]*)"
)


def find_tasks(source_code):
    """Find tasks in source code (TODO, FIXME, XXX, ...)."""
    results = []
    for line, text in enumerate(source_code.splitlines()):
        for todo in re.findall(TASKS_PATTERN, text):
            todo_text = (todo[-1].strip(' :').capitalize() if todo[-1]
                         else todo[-2])
            results.append((todo_text, line + 1))
    return results



# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
# (see spyder/__init__.py for details)
"""
Find in Files Plugin.
"""

# Third party imports
from qtpy.QtCore import Qt
from qtpy.QtWidgets import QApplication

# Local imports
from spyder.api.plugins import Plugins, SpyderDockablePlugin
from spyder.api.plugin_registration.decorators import (
    on_plugin_available, on_plugin_teardown)
from spyder.api.translations import _
from spyder.plugins.findinfiles.widgets.main_widget import FindInFilesWidget
from spyder.plugins.mainmenu.api import ApplicationMenus, SearchMenuSections
from spyder.utils.misc import getcwd_or_home


# --- Constants
# ----------------------------------------------------------------------------
class FindInFilesActions:
    FindInFiles = 'find in files'


# --- Plugin
# ----------------------------------------------------------------------------
class FindInFiles(SpyderDockablePlugin):
    """
    Find in files DockWidget.
    """
    NAME = 'find_in_files'
    REQUIRES = []
    OPTIONAL = [
        Plugins.Editor,
        Plugins.Projects,
        Plugins.MainMenu,
        Plugins.WorkingDirectory,
    ]
    TABIFY = [Plugins.VariableExplorer]
    WIDGET_CLASS = FindInFilesWidget
    CONF_SECTION = NAME
    CONF_FILE = False
    RAISE_AND_FOCUS = True

    # --- SpyderDocakblePlugin API
    # ------------------------------------------------------------------------
    @staticmethod
    def get_name():
        return _("Find")

    @staticmethod
    def get_description():
        return _("Search for text patterns in files.")

    @classmethod
    def get_icon(cls):
        return cls.create_icon('findf')

    def on_initialize(self):
        self.create_action(
            FindInFilesActions.FindInFiles,
            text=_("Search text in files..."),
            tip=_("Search text in multiple files with the Find pane"),
            triggered=self.find,
            register_shortcut=True,
            context=Qt.WindowShortcut
        )
        self.refresh_search_directory()

    @on_plugin_available(plugin=Plugins.Editor)
    def on_editor_available(self):
        widget = self.get_widget()
        editor = self.get_plugin(Plugins.Editor)
        widget.sig_edit_goto_requested.connect(
            lambda filename, lineno, search_text, colno, colend: editor.load(
                filename, lineno, start_column=colno, end_column=colend))
        editor.sig_file_opened_closed_or_updated.connect(
            self.set_current_opened_file)

    @on_plugin_available(plugin=Plugins.Projects)
    def on_projects_available(self):
        projects = self.get_plugin(Plugins.Projects)
        projects.sig_project_loaded.connect(self.set_project_path)
        projects.sig_project_closed.connect(self.unset_project_path)

    @on_plugin_available(plugin=Plugins.MainMenu)
    def on_main_menu_available(self):
        mainmenu = self.get_plugin(Plugins.MainMenu)
        findinfiles_action = self.get_action(FindInFilesActions.FindInFiles)

        mainmenu.add_item_to_application_menu(
            findinfiles_action,
            menu_id=ApplicationMenus.Search,
            section=SearchMenuSections.FindInFiles
        )

    @on_plugin_available(plugin=Plugins.WorkingDirectory)
    def on_working_directory_available(self):
        working_directory = self.get_plugin(Plugins.WorkingDirectory)
        working_directory.sig_current_directory_changed.connect(
            self.refresh_search_directory
        )

    @on_plugin_teardown(plugin=Plugins.Editor)
    def on_editor_teardown(self):
        widget = self.get_widget()
        editor = self.get_plugin(Plugins.Editor)
        widget.sig_edit_goto_requested.disconnect()
        editor.sig_file_opened_closed_or_updated.disconnect(
            self.set_current_opened_file)

    @on_plugin_teardown(plugin=Plugins.Projects)
    def on_projects_teardon_plugin_teardown(self):
        projects = self.get_plugin(Plugins.Projects)
        projects.sig_project_loaded.disconnect(self.set_project_path)
        projects.sig_project_closed.disconnect(self.unset_project_path)

    @on_plugin_teardown(plugin=Plugins.MainMenu)
    def on_main_menu_teardown(self):
        mainmenu = self.get_plugin(Plugins.MainMenu)

        mainmenu.remove_item_from_application_menu(
            FindInFilesActions.FindInFiles,
            menu_id=ApplicationMenus.Search,
        )

    @on_plugin_teardown(plugin=Plugins.WorkingDirectory)
    def on_working_directory_teardown(self):
        working_directory = self.get_plugin(Plugins.WorkingDirectory)
        working_directory.sig_current_directory_changed.disconnect(
            self.refresh_search_directory
        )

    def on_close(self, cancelable=False):
        self.get_widget()._update_options()
        if self.get_widget().running:
            self.get_widget()._stop_and_reset_thread(ignore_results=True)
        return True

    # --- Public API
    # ------------------------------------------------------------------------
    def refresh_search_directory(self):
        """
        Refresh search directory.
        """
        self.get_widget().set_directory(getcwd_or_home())

    def set_current_opened_file(self, path, _language):
        """
        Set path of current opened file in editor.

        Parameters
        ----------
        path: str
            Path of editor file.
        """
        self.get_widget().set_file_path(path)

    def set_project_path(self, path):
        """
        Set and refresh current project path.

        Parameters
        ----------
        path: str
            Opened project path.
        """
        self.get_widget().set_project_path(path)

    def set_max_results(self, value=None):
        """
        Set maximum amount of results to add to the result browser.

        Parameters
        ----------
        value: int, optional
            Number of results. If None an input dialog will be used.
            Default is None.
        """
        self.get_widget().set_max_results(value)

    def unset_project_path(self):
        """
        Unset current project path.
        """
        self.get_widget().disable_project_search()

    def find(self):
        """
        Search text in multiple files.

        Notes
        -----
        Find in files using the currently selected text of the focused widget.
        """
        focus_widget = QApplication.focusWidget()
        text = ''
        try:
            if focus_widget.has_selected_text():
                text = focus_widget.get_selected_text()
        except AttributeError:
            # This is not a text widget deriving from TextEditBaseWidget
            pass

        self.switch_to_plugin()
        widget = self.get_widget()

        if text:
            widget.set_search_text(text)

        widget.find()


def test():
    import sys

    from spyder.config.manager import CONF
    from spyder.utils.qthelpers import qapplication

    app = qapplication()
    widget = FindInFiles(None, CONF)
    widget.show()
    sys.exit(app.exec_())


if __name__ == '__main__':
    test()



# -*- coding: utf-8 -*-
#
# Copyright © 2025 Spyder Project Contributors
# Licensed under the terms of the MIT License
# (see spyder/__init__.py for details)

from __future__ import annotations

from http import HTTPStatus

import aiohttp

from spyder.plugins.remoteclient import SPYDER_PLUGIN_NAME
from spyder.plugins.remoteclient.api.manager.base import (
    SpyderRemoteAPIManagerBase,
)
from spyder.plugins.remoteclient.api.modules.base import (
    SpyderBaseJupyterAPI,
    SpyderRemoteAPIError,
)


@SpyderRemoteAPIManagerBase.register_api
class SpyderRemoteEnvironAPI(SpyderBaseJupyterAPI):
    """
    API for managing the environment variables on the remote server.

    Raises
    ------
    SpyderRemoteAPIError
        If the API call fails.
    aiohttp.ClientResponseError
        If the API call fails with a client error.
    """

    base_url = SPYDER_PLUGIN_NAME + "/environ"

    async def _raise_for_status(self, response: aiohttp.ClientResponse):
        return response.raise_for_status()

    async def get(self, name: str, default: str | None = None) -> str | None:
        """
        Get the environment variable value for the given name.

        Parameters
        ----------
        name : str
            The name of the environment variable.

        Returns
        -------
        str
            The value of the environment variable.

        Raises
        ------
        SpyderRemoteAPIError
            If the API call fails.
        """
        try:
            async with self.session.get(self.api_url / name) as response:
                return await response.text()
        except aiohttp.ClientResponseError as e:
            if e.status == HTTPStatus.NOT_FOUND:
                return default
            msg = f"Failed to get environment variable '{name}': {e}"
            raise SpyderRemoteAPIError(msg) from e

    async def set(self, name: str, value: str) -> None:
        """
        Set the environment variable value for the given name.

        Parameters
        ----------
        name : str
            The name of the environment variable.
        value : str
            The value of the environment variable.
        """
        await self.session.post(self.api_url / name, data={"value": value})

    async def delete(self, name: str) -> None:
        """
        Delete the environment variable for the given name.

        Parameters
        ----------
        name : str
            The name of the environment variable.
        """
        await self.session.delete(self.api_url / name)

    async def to_dict(self) -> dict[str, str]:
        """
        Get the environment variables as a dictionary.

        Returns
        -------
        dict[str, str]
            The environment variables as a dictionary.
        """
        async with self.session.get(self.api_url) as response:
            return await response.json()



# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
# (see spyder/__init__.py for details)

"""
spyder.app
==========

Modules related to starting and restarting the Spyder application
"""
import os
import sys

# On macOS conda installations, sys.executable may be a symlink in the
# application bundle, and therefore should be resolved to the executable in the
# environment. Also store the orignal sys.executable.
SHORTCUT_EXE = None
if sys.platform == "darwin" and sys.executable.endswith("MacOS/python"):
    SHORTCUT_EXE = sys.executable
    sys.executable = os.readlink(sys.executable)



# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
# (see spyder/__init__.py for details)

"""
Debugger Plugin API.
"""

from spyder.plugins.debugger.widgets.main_widget import (
    DebuggerWidgetActions, DebuggerBreakpointActions,
    DebuggerWidgetOptionsMenuSections, DebuggerWidgetMainToolBarSections,
    DebuggerWidgetMenus, DebuggerContextMenuSections
)



# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# based on pylintgui.py by Pierre Raybaut
#
# Licensed under the terms of the MIT License
# (see spyder/__init__.py for details)

"""
Profiler widget.

See the official documentation on python profiling:
https://docs.python.org/3/library/profile.html
"""

# Standard library imports
from collections.abc import Callable
import os
import os.path as osp
import sys
import tempfile
import textwrap

# Third party imports
from qtpy import PYSIDE2
from qtpy.QtCore import Qt, Signal
from qtpy.QtGui import QColor
from qtpy.QtWidgets import (
    QMessageBox,
    QTreeWidget,
    QTreeWidgetItem,
    QVBoxLayout,
    QWidget,
)

# Local imports
from spyder.api.config.mixins import SpyderConfigurationAccessor
from spyder.api.shellconnect.mixins import ShellConnectWidgetForStackMixin
from spyder.api.translations import _
from spyder.api.widgets.mixins import SpyderWidgetMixin
from spyder.utils.icon_manager import ima
from spyder.utils.palette import SpyderPalette
from spyder.utils.qthelpers import set_item_user_text
from spyder.widgets.helperwidgets import FinderWidget


class ProfilerKey:
    """
    Class to save the indexes of the profiler keys

    The quantities calculated by the profiler are as follows
    (from profile.Profile):
    [0] = The number of times this function was called, not counting direct
          or indirect recursion,
    [1] = Number of times this function appears on the stack, minus one
    [2] = Total time spent internal to this function
    [3] = Cumulative time that this function was present on the stack.  In
          non-recursive functions, this is the total execution time from start
          to finish of each invocation of a function, including time spent in
          all subfunctions.
    [4] = A dictionary indicating for each function name, the number of times
          it was called by us.
    """
    Calls = 0
    TotalCalls = 1
    LocalTime = 2
    TotalTime = 3
    Callers = 4


class ProfilerSubWidget(
    QWidget, SpyderWidgetMixin, ShellConnectWidgetForStackMixin
):
    """Profiler widget for shellwidget"""

    # Signals
    sig_display_requested = Signal(object)
    sig_hide_finder_requested = Signal()
    sig_refresh = Signal()

    def __init__(self, parent=None):
        if not PYSIDE2:
            super().__init__(parent, class_parent=parent)
        else:
            QWidget.__init__(self, parent)
            SpyderWidgetMixin.__init__(self, class_parent=parent)

        self.data_tree: ProfilerDataTree | None = None
        self.finder: FinderWidget | None = None
        self.is_profiling = False
        self.recreate_custom_view = False
        self.on_kernel_ready_callback: Callable | None = None

        self.setup()

    # ---- Public API
    # -------------------------------------------------------------------------
    def toggle_finder(self, show):
        """Show and hide the finder."""
        if self.finder is None:
            return
        self.finder.set_visible(show)
        if not show:
            self.data_tree.setFocus()
            self._reset()

    def do_find(self, text):
        """Search for text."""
        if self.data_tree is not None:
            if text:
                self.data_tree.do_find(text)
            else:
                self._reset()

    def finder_is_visible(self):
        """Check if the finder is visible."""
        if self.finder is None:
            return False
        return self.finder.isVisible()

    def finder_text(self):
        return self.finder.text()

    def setup(self):
        """Setup widget."""
        self.data_tree = ProfilerDataTree(self)
        self.data_tree.sig_refresh.connect(self.sig_refresh)
        self._bind_data_tree_methods()

        self.finder = FinderWidget(self)
        self.finder.setVisible(False)
        self.finder.sig_find_text.connect(self.do_find)
        self.finder.sig_hide_finder_requested.connect(
            self.sig_hide_finder_requested
        )
        self.finder.sig_text_cleared.connect(self._reset)

        # Setup layout.
        layout = QVBoxLayout()
        layout.setContentsMargins(0, 0, 0, 0)
        layout.setSpacing(0)
        layout.addWidget(self.data_tree)
        layout.addWidget(self.finder)
        self.setLayout(layout)

    def set_pane_empty(self, empty):
        if empty:
            self.is_empty = True
            self.sig_show_empty_message_requested.emit(True)
        else:
            self.is_empty = False
            self.sig_show_empty_message_requested.emit(False)

    def show_profile_buffer(self, prof_buffer, lib_pathlist):
        """Show profile file."""
        if not prof_buffer:
            return

        # If we're going to show results, profiling has stopped
        self.is_profiling = False

        tmp_dir = None
        if sys.platform.startswith('openbsd'):
            # Do not use /tmp for temporary files
            try:
                from xdg.BaseDirectory import xdg_data_home
                tmp_dir = osp.join(xdg_data_home, "spyder")
                os.makedirs(tmp_dir, exist_ok=True)
            except Exception:
                tmp_dir = None

        with tempfile.TemporaryDirectory(dir=tmp_dir) as tempdir:
            filename = os.path.join(tempdir, "profile.prof")
            with open(filename, "bw") as f:
                f.write(prof_buffer)
            self.data_tree.lib_pathlist = lib_pathlist
            self.data_tree.load_data(filename)

        # Show
        self.set_pane_empty(False)
        self.data_tree._show_tree()
        self.sig_display_requested.emit(self)

    def set_context_menu(self, menu):
        self.data_tree.menu = menu

    # ---- ProfilerDataTree API
    # -------------------------------------------------------------------------
    @property
    def inverted_tree(self):
        return self.data_tree.inverted_tree

    @inverted_tree.setter
    def inverted_tree(self, state):
        self.data_tree.inverted_tree = state

    @property
    def callers_or_callees_enabled(self):
        return self.data_tree.callers_or_callees_enabled

    @callers_or_callees_enabled.setter
    def callers_or_callees_enabled(self, state):
        self.data_tree.callers_or_callees_enabled = state

    @property
    def ignore_builtins(self):
        return self.data_tree.ignore_builtins

    @ignore_builtins.setter
    def ignore_builtins(self, state):
        self.data_tree.ignore_builtins = state

    @property
    def show_slow(self):
        return self.data_tree.show_slow

    @show_slow.setter
    def show_slow(self, state):
        self.data_tree.show_slow = state

    @property
    def profdata(self):
        return self.data_tree.profdata

    @property
    def compare_data(self):
        return self.data_tree.compare_data

    # ---- Private API
    # -------------------------------------------------------------------------
    def _bind_data_tree_methods(self):
        """
        Bind some methods from ProfilerDataTree to this widget so they can be
        easily called in ProfilerWidget.
        """
        for method in [
            "refresh_tree",
            "home_tree",
            "change_view",
            "show_slow_items",
            "undo",
            "redo",
            "show_selected",
            "currentItem",
            "save_data",
            "compare",
        ]:
            setattr(self, method, getattr(self.data_tree, method))

    def _reset(self):
        """Reset view to its initial state."""
        if self.data_tree.show_slow:
            self.data_tree.show_slow_items()
        else:
            self.data_tree._show_tree()


class TreeWidgetItem(QTreeWidgetItem):
    """Item to show in the tree. It represent a function call."""
    def __init__(self, parent, item_key, profile_data, compare_data, icon_list,
                 index_dict):
        QTreeWidgetItem.__init__(self, parent)

        self.item_key = item_key
        self.index_dict = index_dict

        # Order is from profile data
        self.total_calls, self.local_time, self.total_time = profile_data[1:4]
        (
            filename,
            line_number,
            function_name,
            file_and_line,
            node_type,
        ) = self.function_info(item_key)

        self.function_name = function_name
        self.filename = filename
        self.line_number = line_number

        self.set_item_data(filename, line_number)
        self.setIcon(self.index_dict["function_name"], icon_list[node_type])
        self.set_tooltips()

        data = {
            "function_name": function_name,
            "total_time": self.format_measure(self.total_time),
            "local_time": self.format_measure(self.local_time),
            "number_calls": self.format_measure(self.total_calls),
            "file:line": file_and_line
        }
        self.set_data(data)
        alignment = {
            "total_time": Qt.AlignRight,
            "local_time": Qt.AlignRight,
            "number_calls": Qt.AlignRight,
        }
        self.set_alignment(alignment)

        if self.is_recursive():
            self.setData(
                self.index_dict["file:line"],
                Qt.DisplayRole,
                "(%s)" % _("recursion"),
            )
            self.setDisabled(True)

        if compare_data is None:
            return

        diff_data = {}
        diff_colors = {}
        # Keep same order as profile data
        compare_keys = [
            "unused",
            "number_calls_diff",
            "local_time_diff",
            "total_time_diff"
        ]
        for i in range(1, 4):
            diff_str, color = self.color_diff(
                profile_data[i] - compare_data[i]
            )
            diff_data[compare_keys[i]] = diff_str
            diff_colors[compare_keys[i]] = color

        self.set_data(diff_data)
        self.set_color(diff_colors)
        diff_alignment = {
            "total_time_diff": Qt.AlignLeft,
            "local_time_diff": Qt.AlignLeft,
            "number_calls_diff": Qt.AlignLeft
        }
        self.set_alignment(diff_alignment)

    def set_data(self, data):
        """Set data in columns."""
        for k, v in data.items():
            self.setData(self.index_dict[k], Qt.DisplayRole, v)

    def set_color(self, colors):
        """Set colors"""
        for k, v in colors.items():
            self.setForeground(self.index_dict[k], QColor(v))

    def set_alignment(self, alignment):
        """Set alignment."""
        for k, v in alignment.items():
            self.setTextAlignment(self.index_dict[k], v)

    @staticmethod
    def color_diff(difference):
        """Color difference."""
        diff_str = ""
        color = "black"
        if difference:
            color, sign = (
                (SpyderPalette.COLOR_SUCCESS_1, '-')
                if difference < 0
                else (SpyderPalette.COLOR_ERROR_1, '+')
            )
            diff_str = '{}{}'.format(
                sign, TreeWidgetItem.format_measure(difference)
            )
        return diff_str, color

    @staticmethod
    def format_measure(measure):
        """Get format and units for data coming from profiler task."""
        # Convert to a positive value.
        measure = abs(measure)

        # For number of calls
        if isinstance(measure, int):
            return str(measure)

        # For time measurements
        if 1.e-9 < measure <= 1.e-6:
            measure = u"{0:.2f} ns".format(measure / 1.e-9)
        elif 1.e-6 < measure <= 1.e-3:
            measure = u"{0:.2f} \u03BCs".format(measure / 1.e-6)
        elif 1.e-3 < measure <= 1:
            measure = u"{0:.2f} ms".format(measure / 1.e-3)
        elif 1 < measure <= 60:
            measure = u"{0:.2f} s".format(measure)
        elif 60 < measure <= 3600:
            m, s = divmod(measure, 3600)
            if s > 60:
                m, s = divmod(measure, 60)
                s = str(s).split(".")[-1]
            measure = u"{0:.0f}.{1:.2s} min".format(m, s)
        else:
            h, m = divmod(measure, 3600)
            if m > 60:
                m /= 60
            measure = u"{0:.0f}h:{1:.0f}min".format(h, m)
        return measure

    def __lt__(self, otherItem):
        column = self.treeWidget().sortColumn()
        try:
            if column == self.index_dict["total_time"]:
                return self.total_time > otherItem.total_time
            if column == self.index_dict["local_time"]:
                return self.local_time > otherItem.local_time

            return float(self.text(column)) > float(otherItem.text(column))
        except ValueError:
            return self.text(column) > otherItem.text(column)

    def set_item_data(self, filename, line_number):
        """Set tree item user data: filename (string) and line_number (int)"""
        # separator between filename and linenumber
        SEP = r"<[=]>"
        # (must be improbable as a filename)
        set_item_user_text(self, '%s%s%d' % (filename, SEP, line_number))

    def function_info(self, functionKey):
        """Returns processed information about the function's name and file."""
        node_type = 'function'
        filename, line_number, function_name = functionKey

        if function_name == '<module>':
            module_path, module_name = osp.split(filename)
            node_type = 'module'
            if module_name == '__init__.py':
                module_path, module_name = osp.split(module_path)
            function_name = '<' + module_name + '>'

        if not filename or filename == '~':
            file_and_line = '(built-in)'
            node_type = 'builtin'
        else:
            if function_name == '__init__':
                node_type = 'constructor'
            file_and_line = '%s : %d' % (filename, line_number)

        return filename, line_number, function_name, file_and_line, node_type

    def is_recursive(self):
        """Returns True is a function is a descendant of itself."""
        ancestor = self.parent()
        while ancestor:
            if (
                self.function_name == ancestor.function_name
                and self.filename == ancestor.filename
                and self.line_number == ancestor.line_number
            ):
                return True
            else:
                ancestor = ancestor.parent()
        return False

    def set_tooltips(self):
        """Set item tooltips."""
        self.setToolTip(self.index_dict["function_name"], self.function_name)

        if not self.filename or self.filename == '~':
            fname_tip = "(built-in)"
        else:
            fname_tip = f"{self.filename}:{self.line_number}"

        self.setToolTip(self.index_dict["file:line"], fname_tip)


class ProfilerDataTree(QTreeWidget, SpyderConfigurationAccessor):
    """
    Convenience tree widget (with built-in model)
    to store and view profiler data.

    The quantities calculated by the profiler are as follows
    (from profile.Profile):
    [0] = The number of times this function was called, not counting direct
          or indirect recursion,
    [1] = Number of times this function appears on the stack, minus one
    [2] = Total time spent internal to this function
    [3] = Cumulative time that this function was present on the stack.  In
          non-recursive functions, this is the total execution time from start
          to finish of each invocation of a function, including time spent in
          all subfunctions.
    [4] = A dictionary indicating for each function name, the number of times
          it was called by us.
    """

    CONF_SECTION = 'profiler'

    # List of internal functions to exclude
    FUNCTIONS_TO_EXCLUDE = [
        ('~', 0, "<method 'disable' of '_lsprof.Profiler' objects>")
    ]

    # Signals
    sig_refresh = Signal()

    def __init__(self, parent=None):
        if not PYSIDE2:
            super().__init__(parent)
        else:
            QTreeWidget.__init__(self, parent)

        self.header_list = [
            _("Function/Module"),
            _("Total Time"),
            _("Diff"),
            _("Local Time"),
            _("Diff"),
            _("Calls"),
            _("Diff"),
            _("File:line"),
        ]
        self.icon_list = {
            'module': parent.create_icon('python'),
            'function': parent.create_icon('function'),
            'builtin': parent.create_icon('python'),
            'constructor': parent.create_icon('class')
        }
        self.index_dict = {
            "function_name": 0,
            "total_time": 1,
            "total_time_diff": 2,
            "local_time": 3,
            "local_time_diff": 4,
            "number_calls": 5,
            "number_calls_diff": 6,
            "file:line": 7
        }
        self.profdata = None   # To be filled by self.load_data()
        self.items_to_be_shown = None
        self.current_view_depth = None
        self.compare_data = None
        self.inverted_tree = False
        self.callers_or_callees_enabled = False
        self.ignore_builtins = False
        self.show_slow = False
        self.root_key = None
        self.menu = None
        self._last_children = None
        self.setColumnCount(len(self.header_list))
        self.setHeaderLabels(self.header_list)
        self.initialize_view()
        self.itemExpanded.connect(self.item_expanded)
        self.lib_pathlist = None
        self.history = []
        self.redo_history = []

        self.set_tooltips()

    def contextMenuEvent(self, event):
        """Reimplement Qt method"""
        if self.menu is None:
            return
        if self.profdata and self.indexAt(event.pos()).isValid():
            self.menu.popup(event.globalPos())
            event.accept()

    def initialize_view(self):
        """Clean the tree and view parameters"""
        self.clear()
        self.items_to_be_shown = {}
        self.current_view_depth = 0
        if (
            self.compare_data is not None
            and self.compare_data is not self.profdata
        ):
            self.hide_diff_cols(False)
        else:
            self.hide_diff_cols(True)

    def load_data(self, profdatafile):
        """Load profiler data saved by profile/cProfile module"""
        self.history = []
        self.redo_history = []
        if not os.path.isfile(profdatafile):
            self.profdata = None
            return
        import pstats

        # Fixes spyder-ide/spyder#6220.
        try:
            self.profdata = pstats.Stats(profdatafile)
            self.profdata.calc_callees()
            self.root_key = self.find_root()
        except OSError:
            self.profdata = None
            return

    def compare(self, filename):
        """Load compare file."""
        if filename is None:
            self.compare_data = None
            return
        import pstats

        # Fixes spyder-ide/spyder#5587.
        try:
            self.compare_data = pstats.Stats(filename)
            self.compare_data.calc_callees()
            if self.profdata is None:
                # Show the compare data as prof_data
                self.profdata = self.compare_data
                self.root_key = self.find_root()
        except OSError as e:
            QMessageBox.critical(
                self,
                _("Error"),
                _(
                    "Error when trying to load profiler results. The error "
                    "was<br><br>"
                    "<tt>{0}</tt>"
                ).format(e),
            )
            self.compare_data = None

    def hide_diff_cols(self, hide):
        """Hide difference columns."""
        for i in (
            self.index_dict["total_time_diff"],
            self.index_dict["local_time_diff"],
            self.index_dict["number_calls_diff"],
        ):
            self.setColumnHidden(i, hide)

    def save_data(self, filename):
        """Save profiler data."""
        self.profdata.dump_stats(filename)

    def find_root(self):
        """Find a function without a caller."""
        # Fixes spyder-ide/spyder#8336.
        if self.profdata is not None:
            self.profdata.sort_stats("cumulative")
        else:
            return
        for func in self.profdata.fcn_list:
            if (
                ('~', 0) != func[0:2]
                and not func[2].startswith('<built-in method exec>')
            ):
                # This skips the profiler function at the top of the list
                # it does only occur in Python 3
                return func

    def is_builtin(self, key):
        """Check if key is buit-in."""
        path = key[0]
        if not path:
            return True
        if path == "~":
            return True
        if path.startswith("<"):
            return True

        path = os.path.normcase(os.path.normpath(path))
        if self.lib_pathlist is not None:
            for libpath in self.lib_pathlist:
                libpath = os.path.normcase(os.path.normpath(libpath))
                commonpath = os.path.commonpath([libpath, path])
                if libpath == commonpath:
                    return True

        return False

    def find_children(self, parent):
        """Find all functions called by (parent) function."""
        if self.inverted_tree:
            # Return callers
            return self.profdata.stats[parent][ProfilerKey.Callers]
        else:
            # Return callees
            callees = self.profdata.all_callees[parent]
            if self.ignore_builtins:
                callees = [c for c in callees if not self.is_builtin(c)]
            return callees

    def do_find(self, text):
        """Find all function that match text."""
        if self.profdata is None:
            # Nothing to show
            return

        if self.show_slow:
            children = self.get_slow_items()
            children = [c for c in children if text in c[-1]]
            self.show_slow_items(children)
        else:
            children = self.profdata.fcn_list
            children = [c for c in children if text in c[-1]]
            self._show_tree(children)

    def get_slow_items(self):
        """Get items with large local time."""
        children = self.profdata.fcn_list

        # Get slow items
        children = sorted(
            children,
            key=lambda item: self.profdata.stats[item][ProfilerKey.LocalTime],
            reverse=True
        )

        # Ignore builtins
        if self.ignore_builtins:
            children = [c for c in children if not self.is_builtin(c)]

        # Only keep top n_slow_children
        n_children = self.get_conf('n_slow_children')
        return children[:n_children]

    def show_slow_items(self, children=None):
        """Show slow items."""
        if self.profdata is None:
            # Nothing to show
            return

        if children is None:
            children = self.get_slow_items()

        n_children = self.get_conf('n_slow_children')
        self._show_tree(children, max_items=n_children, sort_time="local_time")

    def refresh_tree(self):
        """Refresh tree."""
        self._show_tree(self._last_children)

    def home_tree(self):
        """Reset tree."""
        self._show_tree()

    def show_selected(self):
        """Show current item."""
        self.callers_or_callees_enabled = True
        self._show_tree([self.currentItem().item_key])

    def undo(self):
        """Undo change."""
        if len(self.history) > 1:
            self.redo_history.append(self.history.pop(-1))
            self._show_tree(self.history.pop(-1), reset_redo=False)

    def redo(self):
        """Redo changes."""
        if len(self.redo_history) > 0:
            self._show_tree(self.redo_history.pop(-1), reset_redo=False)

    def _show_tree(
        self,
        children=None,
        max_items=None,
        reset_redo=True,
        sort_time="total_time",
    ):
        """Populate the tree with profiler data and display it."""
        if self.profdata is None:
            # Nothing to show
            return

        self._last_children = children

        # List of frames to hide at the top
        head_list = [self.root_key, ]
        head_list += list(
            self.profdata.stats[self.root_key][ProfilerKey.Callers])

        if children is None:
            if self.inverted_tree:
                # Show all callees
                self.tree_state = None
                children = []
                for key, value in self.profdata.all_callees.items():
                    if key in self.FUNCTIONS_TO_EXCLUDE or key in head_list:
                        continue
                    if self.ignore_builtins:
                        if not self.is_builtin(key):
                            non_builtin_callees = [
                                k for k in value if not self.is_builtin(k)
                            ]
                            if len(non_builtin_callees) == 0:
                                children.append(key)
                    else:
                        if len(value) == 0:
                            children.append(key)
            else:
                # Show all called
                self.tree_state = None
                rootkey = self.root_key  # This root contains profiler overhead
                if rootkey is not None:
                    children = self.find_children(rootkey)
        else:
            if self.ignore_builtins:
                children = [c for c in children if not self.is_builtin(c)]
            children = [
                c for c in children if c not in self.FUNCTIONS_TO_EXCLUDE
            ]
            if max_items is not None:
                children = children[:max_items]

        self.initialize_view()  # Clear before re-populating
        self.setItemsExpandable(True)
        self.setSortingEnabled(False)
        if children is not None:
            if len(self.history) == 0 or self.history[-1] != children:
                # Do not add twice the same element
                self.history.append(children)
                if reset_redo:
                    self.redo_history = []

            # Populate the tree
            self.populate_tree(self, children)
            self.setSortingEnabled(True)
            self.sortItems(self.index_dict[sort_time], Qt.AscendingOrder)
            self.resizeColumnToContents(0)

        self.sig_refresh.emit()

    def populate_tree(self, parentItem, children_list):
        """
        Recursive method to create each item (and associated data)
        in the tree.
        """
        children_list = [
            c for c in children_list if c not in self.FUNCTIONS_TO_EXCLUDE
        ]

        for child_key in children_list:
            item_profdata, item_compdata = self.get_item_data(child_key)
            child_item = TreeWidgetItem(
                parentItem,
                child_key,
                item_profdata,
                item_compdata,
                self.icon_list,
                self.index_dict
            )
            if not child_item.is_recursive():
                grandchildren_list = self.find_children(child_key)
                if grandchildren_list:
                    child_item.setChildIndicatorPolicy(
                        child_item.ShowIndicator
                    )
                    self.items_to_be_shown[id(child_item)] = grandchildren_list

    def get_item_data(self, item_key):
        """Return the profile and compare data for the item_key."""
        item_profdata = self.profdata.stats.get(item_key, [0, 0, 0, 0, {}])
        item_compdata = None
        if self.compare_data is not None:
            item_compdata = self.compare_data.stats.get(
                item_key, [0, 0, 0, 0, {}]
            )
        return item_profdata, item_compdata

    def item_expanded(self, item):
        """Fill item children."""
        if item.childCount() == 0 and id(item) in self.items_to_be_shown:
            children_list = self.items_to_be_shown[id(item)]
            self.populate_tree(item, children_list)

    def get_top_level_items(self):
        """Iterate over top level items."""
        return [self.topLevelItem(_i)
                for _i in range(self.topLevelItemCount())]

    def get_items(self, maxlevel):
        """Return all items with a level <= `maxlevel`"""
        itemlist = []

        def add_to_itemlist(item, maxlevel, level=1):
            level += 1
            for index in range(item.childCount()):
                citem = item.child(index)
                itemlist.append(citem)
                if level <= maxlevel:
                    add_to_itemlist(citem, maxlevel, level)

        for tlitem in self.get_top_level_items():
            itemlist.append(tlitem)
            if maxlevel > 0:
                add_to_itemlist(tlitem, maxlevel=maxlevel)

        return itemlist

    def change_view(self, change_in_depth):
        """
        Change the view depth by expand or collapsing all same-level nodes.
        """
        self.current_view_depth += change_in_depth
        if self.current_view_depth < 0:
            self.current_view_depth = 0
        self.collapseAll()
        if self.current_view_depth > 0:
            for item in self.get_items(maxlevel=self.current_view_depth-1):
                item.setExpanded(True)

    def set_tooltips(self):
        """Set tooltips."""
        tooltips = {
            "function_name": _('Function or module name'),
            "total_time": _(
                'Time spent in function (including sub-functions)'
            ),
            "local_time": _(
                'Local time spent in function (not in sub-functions)'
            ),
            "number_calls": _('Total number of calls (including recursion)'),
            "file:line": _('File and line where the function is defined')
        }

        for column_name, tip_text in tooltips.items():
            self.headerItem().setIcon(
                self.index_dict[column_name], ima.icon('question_tip_hover')
            )

            tip_text = '\n'.join(textwrap.wrap(tip_text, 50))
            self.headerItem().setToolTip(
                self.index_dict[column_name], tip_text
            )



# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
# (see spyder/__init__.py for details)

"""
This module contains the Line Number panel
"""

# Standard library imports
from math import ceil
import math

# Third party imports
from qtpy.QtCore import QSize, Qt, QPointF
from qtpy.QtGui import (
    QColor, QFont, QFontMetricsF, QPainter, QStaticText, QTextOption
)

# Local imports
from spyder.utils.icon_manager import ima
from spyder.plugins.editor.api.panel import Panel
from spyder.plugins.completion.api import DiagnosticSeverity


class LineNumberArea(Panel):
    """Line number area (on the left side of the text editor widget)"""

    # --- Qt Overrides
    # -----------------------------------------------------------------

    def __init__(self):
        Panel.__init__(self)

        self.setMouseTracking(True)
        self.scrollable = True
        self.linenumbers_color = QColor(Qt.darkGray)

        # Markers
        self._markers_margin = True

        # Icons
        self.error_icon = ima.icon('error')
        self.warning_icon = ima.icon('warning')
        self.info_icon = ima.icon('information')
        self.hint_icon = ima.icon('hint')
        self.todo_icon = ima.icon('todo')

        # Line number area management
        self._margin = True
        self._pressed = -1
        self._released = -1

        # This is a tuple composed of (number of digits, current width)
        self._width_cache = None

        # Cache line numbers
        self._static_line_numbers = None
        self._static_active_line = None

        # Static text must be flushed when dpi changes (qt bug?)
        self._static_text_dpi = None

    def sizeHint(self):
        """Override Qt method."""
        return QSize(self.compute_width(), 0)

    def paintEvent(self, event):
        """Override Qt method.

        Painting line number area
        """

        painter = QPainter(self)
        painter.fillRect(event.rect(), self.editor.sideareas_color)
        font_height = self.editor.fontMetrics().height()

        def draw_pixmap(xleft, ytop, pixmap):
            # Scale pixmap height to device independent pixels
            pixmap_height = pixmap.height() / pixmap.devicePixelRatio()
            painter.drawPixmap(
                xleft,
                ceil(ytop + (font_height-pixmap_height) / 2),
                pixmap
            )

        size = self.get_markers_margin() - 2
        icon_size = QSize(size, size)

        if self._margin:
            font = self.editor.font()
            fm = QFontMetricsF(font)
            if (
                fm.leading() == 0
                and self.editor.lineWrapMode() == QTextOption.NoWrap
            ):
                self.draw_linenumbers(painter)
            else:
                # The editor doesn't care about leading or the text is being
                # wrapped, so each line must be drawn independently.
                self.draw_linenumbers_slow(painter)
        self.paint_cell(painter)

        for top, line_number, block in self.editor.visible_blocks:
            data = block.userData()
            if data:
                if data.code_analysis:
                    errors = 0
                    warnings = 0
                    infos = 0
                    hints = 0
                    for _, _, sev, _ in data.code_analysis:
                        errors += sev == DiagnosticSeverity.ERROR
                        warnings += sev == DiagnosticSeverity.WARNING
                        infos += sev == DiagnosticSeverity.INFORMATION
                        hints += sev == DiagnosticSeverity.HINT

                    if errors:
                        draw_pixmap(1, top, self.error_icon.pixmap(icon_size))
                    elif warnings:
                        draw_pixmap(
                            1, top, self.warning_icon.pixmap(icon_size))
                    elif infos:
                        draw_pixmap(1, top, self.info_icon.pixmap(icon_size))
                    elif hints:
                        draw_pixmap(1, top, self.hint_icon.pixmap(icon_size))

                if self._markers_margin and data.todo:
                    draw_pixmap(1, top, self.todo_icon.pixmap(icon_size))

    def draw_linenumbers(self, painter):
        """Draw line numbers."""
        if len(self.editor.visible_blocks) == 0:
            return
        active_line_number = self.editor.textCursor().blockNumber() + 1
        number_digits = self.compute_width_digits()
        width = self.width()

        visible_lines = [ln for _, ln, _ in self.editor.visible_blocks]

        try:
            idx = visible_lines.index(active_line_number)
            active_top = self.editor.visible_blocks[idx][0]
        except ValueError:
            active_top = None

        # Right align
        line_numbers = [f"{ln:{number_digits}d}" for ln in visible_lines]

        # Use non-breaking spaces and <br> returns
        lines = "<br>".join(line_numbers).replace(" ", "&nbsp;")

        # This is needed to make that the font size of line numbers
        # be the same as the text one when zooming
        # See spyder-ide/spyder#2296 and spyder-ide/spyder#4811.
        font = self.editor.font()
        font.setWeight(QFont.Weight.Normal)
        painter.setFont(font)
        painter.setPen(self.linenumbers_color)

        if self.logicalDpiX() != self._static_text_dpi:
            self._static_text_dpi = self.logicalDpiX()
            self._static_line_numbers = None
            self._static_active_line = None

        if self._static_line_numbers:
            if lines != self._static_line_numbers.text():
                self._static_line_numbers.setText(lines)
        else:
            self._static_line_numbers = QStaticText(lines)
            self._static_line_numbers.prepare(font=font)

        top = self.editor.visible_blocks[0][0]
        left = width - self._static_line_numbers.size().width()

        painter.drawStaticText(
            QPointF(left, top), self._static_line_numbers)

        if active_top is not None:
            font.setWeight(QFont.Weight.Bold)
            painter.setFont(font)
            painter.setPen(self.editor.normal_color)

            text = str(active_line_number)
            if self._static_active_line:
                if text != self._static_active_line.text():
                    self._static_active_line.setText(text)
            else:
                self._static_active_line = QStaticText(text)
                self._static_active_line.setTextFormat(Qt.PlainText)
                self._static_active_line.prepare(font=font)

            size = self._static_active_line.size()
            left = width - size.width()

            # Hide non-bold number
            painter.fillRect(
                int(left), active_top, int(size.width()), int(size.height()),
                self.editor.sideareas_color
            )

            # Paint bold number
            painter.drawStaticText(
                QPointF(left, active_top), self._static_active_line)

    def draw_linenumbers_slow(self, painter):
        """
        Slower way (2x) to draw line numbers.

        This is necessary for some fonts and when the wrap lines option
        is active.
        """
        font = self.editor.font()
        font_height = self.editor.fontMetrics().height()
        active_block = self.editor.textCursor().block()
        active_line_number = active_block.blockNumber() + 1
        for top, line_number, block in self.editor.visible_blocks:
            if self._margin:
                if line_number == active_line_number:
                    font.setWeight(QFont.Weight.Bold)
                    painter.setFont(font)
                    painter.setPen(self.editor.normal_color)
                else:
                    font.setWeight(QFont.Weight.Normal)
                    painter.setFont(font)
                    painter.setPen(self.linenumbers_color)

                painter.drawText(0, top, self.width(),
                                 font_height,
                                 int(Qt.AlignRight | Qt.AlignBottom),
                                 str(line_number))

    def leaveEvent(self, event):
        """Override Qt method."""
        self.editor.hide_tooltip()

    def mouseMoveEvent(self, event):
        """Override Qt method.

        Show code analisis, if left button pressed select lines.
        """
        line_number = self.editor.get_linenumber_from_mouse_event(event)
        block = self.editor.document().findBlockByNumber(line_number - 1)
        data = block.userData()

        # This disables messages if there is an active drag/selection operation
        check = self._released == -1
        if check and data:
            if data.code_analysis:
                self.editor.show_code_analysis_results(line_number, data)
            elif data.todo:
                self.editor.show_todo(line_number, data)
        else:
            self.editor.hide_tooltip()

        if event.buttons() == Qt.LeftButton:
            self._released = line_number
            self.editor.select_lines(self._pressed, self._released)

    def mousePressEvent(self, event):
        """Override Qt method

        Select line, and starts selection
        """
        line_number = self.editor.get_linenumber_from_mouse_event(event)
        self._pressed = line_number
        self._released = line_number
        self.editor.select_lines(self._pressed,
                                 self._released)

    def mouseReleaseEvent(self, event):
        """Override Qt method."""
        self._released = -1
        self._pressed = -1

    def wheelEvent(self, event):
        """Override Qt method."""
        self.editor.wheelEvent(event)

    # --- Other methods
    # -----------------------------------------------------------------

    def compute_width_digits(self):
        """Compute and return line number area width in digits."""
        number_lines = self.editor.blockCount()
        return max(1, math.ceil(math.log10(
             number_lines + 1)))

    def compute_width(self):
        """Compute and return line number area width."""
        if not self._enabled:
            return 0
        number_digits = self.compute_width_digits()
        if (self._width_cache is not None and
                self._width_cache[0] == number_digits):
            return self._width_cache[1]

        if self._margin:
            margin = 3 + self.editor.fontMetrics().width('9' * number_digits)
        else:
            margin = 0
        width = margin + self.get_markers_margin()
        self._width_cache = (number_digits, width)
        return width

    def _clear_width_cache(self):
        """Clear width cache."""
        self._width_cache = None

    def on_install(self, editor):
        """Clear width cache on font change."""
        super().on_install(editor)
        editor.sig_font_changed.connect(self._clear_width_cache)

    def on_uninstall(self):
        """Disconnect signal."""
        self.editor.sig_font_changed.disconnect(self._clear_width_cache)
        super().on_uninstall()

    def get_markers_margin(self):
        """Get marker margins."""
        font_height = self.editor.fontMetrics().height() + 2
        return font_height

    def setup_margins(self, linenumbers=True, markers=True):
        """
        Setup margin settings
        (except font, now set in editor.set_font)
        """
        self._width_cache = None
        self._margin = linenumbers
        self._markers_margin = markers
        self.set_enabled(linenumbers or markers)

    def set_enabled(self, state):
        self._enabled = state
        self.setVisible(state)

    def get_width(self):
        """Return current line number area width"""
        return self.contentsRect().width()



# -----------------------------------------------------------------------------
# Copyright (c) 2021- Spyder Project Contributors
#
# Released under the terms of the MIT License
# (see LICENSE.txt in the project root directory for details)
# -----------------------------------------------------------------------------

"""
Main container widget for non-dockable Spyder plugins.

:class:`~spyder.api.plugins.SpyderPluginV2` plugins must provide a
:attr:`~spyder.api.plugins.SpyderPluginV2.CONTAINER_CLASS` attribute that is
a subclass of :class:`PluginMainContainer`, if they have additional widgets
like status bar items or toolbars.
"""

from __future__ import annotations

# Standard library imports
from typing import TYPE_CHECKING

# Third party imports
from qtpy import PYSIDE2
from qtpy.QtCore import Signal
from qtpy.QtWidgets import QWidget

# Local imports
from spyder.api.widgets.mixins import SpyderWidgetMixin

if TYPE_CHECKING:
    from qtpy.QtGui import QCloseEvent

    import spyder.app.mainwindow
    from spyder.api.plugins import SpyderPluginV2


class PluginMainContainer(QWidget, SpyderWidgetMixin):
    """
    Main container widget class for non-dockable Spyder plugins.

    This class is used by non-dockable plugins to be able to contain, parent
    and store references to other widgets, like status bar widgets, toolbars,
    context menus, etc.

    .. important::

        If a Spyder non-dockable plugins defines a
        :attr:`~spyder.api.plugins.SpyderPluginV2.CONTAINER_CLASS`
        it must inherit from this class, :class`PluginMainContainer`.
    """

    CONTEXT_NAME: str | None = None
    """
    The name under which to store actions, toolbars, toolbuttons and menus.

    This optional attribute defines the context name under which actions,
    toolbars, toolbuttons and menus should be registered in the
    Spyder global registry.

    If those elements belong to the global scope of the plugin, then this
    attribute should have a ``None`` value, which will use the plugin's name as
    the context scope.
    """

    # ---- Signals
    # ------------------------------------------------------------------------
    sig_free_memory_requested: Signal = Signal()
    """
    Signal to request the main application garbage-collect deleted objects.
    """

    sig_quit_requested: Signal = Signal()
    """
    Signal to request the main Spyder application quit.
    """

    sig_restart_requested: Signal = Signal()
    """
    Signal to request the main Spyder application quit and restart itself.
    """

    sig_redirect_stdio_requested: Signal = Signal(bool)
    """
    Request the main app redirect standard out/error within file pickers.

    This will redirect :data:`~sys.stdin`, :data:`~sys.stdout`, and
    :data:`~sys.stderr` when using :guilabel:`Open`, :guilabel:`Save`,
    and :guilabel:`Browse` dialogs within a plugin's widgets.

    Parameters
    ----------
    enable: bool
        Enable (``True``) or disable (``False``) standard input/output
        redirection.
    """

    sig_exception_occurred: Signal = Signal(dict)
    """
    Signal to report an exception from a plugin.

    Parameters
    ----------
    error_data: dict[str, str | bool]
        The dictionary containing error data. The expected keys are:

        .. code-block:: python

            error_data = {
                "text": str,
                "is_traceback": bool,
                "repo": str,
                "title": str,
                "label": str,
                "steps": str,
            }

        The ``is_traceback`` key indicates if ``text`` contains plain text or a
        Python error traceback.

        The ``title`` and ``repo`` keys indicate how the error data should
        customize the report dialog and GitHub error submission.

        The ``label`` and ``steps`` keys allow customizing the content of the
        error dialog.
    """

    sig_unmaximize_plugin_requested: Signal = Signal((), (object,))
    """
    Request the main window unmaximize the currently maximized plugin, if any.

    If emitted without arguments, it'll unmaximize any plugin.

    Parameters
    ----------
    plugin_instance: SpyderDockablePlugin
        Unmaximize current plugin only if it is not ``plugin_instance``.
    """

    def __init__(
        self,
        name: str,
        plugin: SpyderPluginV2,
        parent: spyder.app.mainwindow.MainWindow | None = None,
    ) -> None:
        """
        Create a new container class for a plugin.
        
        This method is not meant to be overridden by container subclasses.
        Use the :meth:`setup` method instead to instantiate the widgets that
        this one will contain.

        Parameters
        ----------
        name : str
            The name of the plugin, i.e. the
            :attr:`SpyderPluginV2.NAME <spyder.api.plugins.SpyderPluginV2.NAME>`.
        plugin : SpyderPluginV2
            The plugin object this is to be the container class of.
        parent : spyder.app.mainwindow.MainWindow | None, optional
            The container's parent widget, normally the Spyder main window.
            By default (``None``), no parent widget (used for testing).

        Returns
        -------
        None
        """
        if not PYSIDE2:
            super().__init__(parent=parent, class_parent=plugin)
        else:
            QWidget.__init__(self, parent)
            SpyderWidgetMixin.__init__(self, class_parent=plugin)

        # ---- Attributes
        # --------------------------------------------------------------------
        self._name = name
        self._plugin = plugin
        self._parent = parent

        self.PLUGIN_NAME: str = name
        """
        Plugin name in the action, toolbar, toolbutton & menu registries.

        Usually the same as
        :attr:`SpyderPluginV2.NAME <spyder.api.plugins.SpyderPluginV2.NAME>`,
        but may be different from :attr:`CONTEXT_NAME`.
        """

        # Widget setup
        # A PluginMainContainer inherits from QWidget so it can be a parent
        # for the widgets it contains. Since it is a QWidget it will occupy a
        # physical space on the screen and may cast "shadow" on the top left
        # of the main window. To prevent this we ensure the widget has zero
        # width and zero height.
        # See: spyder-ide/spyder#13547
        self.setMaximumWidth(0)
        self.setMaximumHeight(0)

    # ---- Public Qt overridden methods
    # -------------------------------------------------------------------------
    def closeEvent(self, event: QCloseEvent) -> None:
        """
        Handle closing this container widget.

        Parameters
        ----------
        event : QCloseEvent
            The event object closing this widget.

        Returns
        -------
        None
        """
        self.on_close()
        super().closeEvent(event)

    # ---- API: methods to define or override
    # ------------------------------------------------------------------------
    def setup(self) -> None:
        """
        Create widgets, toolbars and menus, and perform other setup steps.

        Returns
        -------
        None

        Raises
        ------
        NotImplementedError
            If the container subclass doesn't define a ``setup`` method.
        """
        raise NotImplementedError(
            "A PluginMainContainer subclass must define a `setup` method!"
        )

    def update_actions(self) -> None:
        """
        Update the state of exposed actions.

        Exposed actions are actions created by the
        :meth:`~spyder.api.widgets.mixins.SpyderActionMixin.create_action`
        method.

        Returns
        -------
        None

        Raises
        ------
        NotImplementedError
            If the subclass doesn't define an ``update_actions`` method.
        """
        raise NotImplementedError(
            "A PluginMainContainer subclass must define a `update_actions` "
            "method!"
        )

    def on_close(self) -> None:
        """
        Perform actions before the container widget is closed.

        Does nothing by default; intended to be overridden for widgets
        that need to perform actions on close.

        .. warning::

            This method **must** only operate on local attributes.

        Returns
        -------
        None
        """
        pass
